"""
Read Redvox data from a single directory
Data files can be either API 900 or API 1000 data formats
The ReadResult object converts api900 data into api 1000 format
"""
from collections import defaultdict
from multiprocessing import Pool, cpu_count
from typing import List, Optional, Dict
from datetime import timedelta

import numpy as np

from redvox.api1000.wrapped_redvox_packet.wrapped_packet import WrappedRedvoxPacketM
from redvox.common import date_time_utils as dtu
from redvox.common import offset_model
from redvox.common import api_conversions as ac
from redvox.common import io
from redvox.common import file_statistics as fs
import redvox.common.parallel_utils as parallel
from redvox.common.station import Station


DEFAULT_GAP_TIME_S: float = 0.25  # default length of a gap in seconds


class ApiReader:
    """
    Reads data from api 900 or api 1000 format, converting all data read into WrappedRedvoxPacketM for
        ease of comparison and use.
    Properties:
        filter: io.ReadFilter with the station ids, start and end time, start and end time padding, and
                types of files to read
        base_dir: str of the directory containing all the files to read
        structured_dir: bool, if True, the base_dir contains a specific directory structure used by the
                        respective api formats.  If False, base_dir only has the data files.  Default False.
        files_index: io.Index of the files that match the filter that are in base_dir
        files_stats: list of Fs.StationStat for the stations in the filter
        index_summary: io.IndexSummary of the filtered data
        gap_time_s: int, minimum amount of seconds between data points to be considered a gap.  Default GAP_TIME_S
        apply_correction: bool, if True, apply timing correction values based on the data in the packets.  Default True
        debug: bool, if True, output additional information during function execution.  Default False.
    """

    def __init__(
        self,
        base_dir: str,
        structured_dir: bool = False,
        read_filter: io.ReadFilter = None,
        debug: bool = False,
    ):
        """
        Initialize the ApiReader object
        :param base_dir: directory containing the files to read
        :param structured_dir: if True, base_dir contains a specific directory structure used by the respective
                                api formats.  If False, base_dir only has the data files.  Default False.
        :param read_filter: ReadFilter for the data files, if None, get everything.  Default None
        :param debug: if True, output additional statements during function execution.  Default False.
        """
        if read_filter:
            self.filter = read_filter
            if self.filter.station_ids:
                self.filter.station_ids = set(self.filter.station_ids)
        else:
            self.filter = io.ReadFilter()
        self.base_dir = base_dir
        self.structured_dir = structured_dir
        self.debug = debug
        self.files_index = self._get_all_files()
        self.index_summary = io.IndexSummary.from_index(self.files_index)

    def _get_all_files(self) -> io.Index:
        """
        get all files in the base dir of the ApiReader
        :return: index with all the files that match the filter
        """
        index = io.Index()
        # this guarantees that all ids we search for are valid
        all_index = self._apply_filter()
        for station_id in all_index.summarize().station_ids():
            station_filter = self.filter.clone()
            checked_index = self._check_station_stats(station_filter.with_station_ids({station_id}))
            index.append(checked_index.entries)
        return index

    def _apply_filter(self, reader_filter: Optional[io.ReadFilter] = None) -> io.Index:
        """
        apply the filter of the reader, or another filter if specified
        :param reader_filter: optional filter; if None, use the reader's filter, default None
        :return: index of the filtered files
        """
        if not reader_filter:
            reader_filter = self.filter
        if self.structured_dir:
            index = io.index_structured(self.base_dir, reader_filter)
        else:
            index = io.index_unstructured(self.base_dir, reader_filter)
        return index

    def _check_station_stats(self, request_filter: io.ReadFilter) -> io.Index:
        """
        recursively check the filter's results; if resulting index has enough information in it,
        return it, otherwise search for more data.
        The filter should only request one station
        :param request_filter: filter representing the requested information
        :return: Index that includes as much information as possible that fits the request
        """
        index = self._apply_filter(request_filter)
        # if there are no restrictions on time or we found nothing, return the index
        if (not self.filter.start_dt and not self.filter.end_dt) or \
                (not request_filter.start_dt and not request_filter.end_dt) or len(index.entries) < 1:
            return index
        stats = fs.extract_stats(index)
        # punt if duration or other important values are invalid
        if any(st.packet_duration == 0.0 or not st.packet_duration for st in stats):
            return index
        latencies = [
            st.latency
            if st.latency is not None and not np.isnan(st.latency) else np.nan
            for st in stats
        ]
        # no latencies means no correction means we're done
        if len(latencies) < 1:
            return index
        # get the model and calculate the revised start and end offsets
        # best fit would be to use the machine timestamps that correspond to the best latencies,
        #   but for now we use the packet start times + 1/2 the packet duration
        offsets = [st.offset if st.offset is not None and not np.isnan(st.offset) else np.nan for st in stats]
        times = [dtu.datetime_to_epoch_microseconds_utc(st.packet_start_dt) for st in stats]
        packet_duration = np.mean([dtu.seconds_to_microseconds(st.packet_duration.total_seconds()) for st in stats])
        model = offset_model.OffsetModel(np.array(latencies), np.array(offsets),
                                         times + 0.5 * packet_duration, 5, 3, times[0],
                                         times[-1] + packet_duration)
        # revise packet's times to real times and compare to requested values
        start_offset = timedelta(microseconds=model.get_offset_at_new_time(
            dtu.datetime_to_epoch_microseconds_utc(stats[0].packet_start_dt)))
        revised_start = stats[0].packet_start_dt + start_offset
        end = stats[-1].packet_start_dt + stats[-1].packet_duration
        end_offset = timedelta(microseconds=model.get_offset_at_new_time(dtu.datetime_to_epoch_microseconds_utc(end)))
        revised_end = end + end_offset
        # if our filtered files encompass the request even when the packet times are updated, return the index
        if (not self.filter.start_dt or revised_start <= self.filter.start_dt) \
                and (not self.filter.end_dt or revised_end >= self.filter.end_dt):
            return index
        # we have to update our filter to get more information
        new_filter = request_filter.clone()
        no_more_start = True
        no_more_end = True
        # check if there is a packet just beyond the request times
        if self.filter.start_dt and revised_start > self.filter.start_dt:
            beyond_start = self.filter.start_dt - np.abs(start_offset) - stats[0].packet_duration
            start_filter = request_filter.clone().with_start_dt(beyond_start).with_end_dt(stats[0].packet_start_dt) \
                .with_end_dt_buf(timedelta(seconds=0))
            start_index = self._apply_filter(start_filter)
            # if the beyond check produces an earlier start date time,
            #  then update filter, otherwise flag result as no more data to obtain
            if len(start_index.entries) > 0 and start_index.entries[0].date_time < index.entries[0].date_time:
                new_filter.with_start_dt(beyond_start)
                no_more_start = False
        if self.filter.end_dt and revised_end < self.filter.end_dt:
            beyond_end = self.filter.end_dt + np.abs(end_offset)
            end_filter = request_filter.clone().with_start_dt(stats[-1].packet_start_dt + stats[-1].packet_duration)\
                .with_end_dt(beyond_end).with_start_dt_buf(timedelta(seconds=0))
            end_index = self._apply_filter(end_filter)
            # if the beyond check produces a later end date time,
            #  then update filter, otherwise flag result as no more data to obtain
            if len(end_index.entries) > 0 and end_index.entries[-1].date_time > index.entries[-1].date_time:
                new_filter.with_end_dt(beyond_end)
                no_more_end = False
        # if there is no more data to obtain from either end, return the original index
        if no_more_start and no_more_end:
            return index
        # check the updated index
        return self._check_station_stats(new_filter)

    def read_files(self) -> Dict[str, List[WrappedRedvoxPacketM]]:
        """
        read all the files in the index
        :return: dictionary of id: list of WrappedRedvoxPacketM, converted from API 900 if necessary
        """
        result: Dict[str, List[WrappedRedvoxPacketM]] = defaultdict(list)

        # Iterate over the API 900 packets in a memory efficient way
        # and convert to API 1000
        # noinspection PyTypeChecker
        for packet_900 in self.files_index.stream(
            io.ReadFilter.empty().with_api_versions({io.ApiVersion.API_900})
        ):
            # noinspection Mypy
            result[packet_900.redvox_id()].append(
                ac.convert_api_900_to_1000(packet_900)
            )

        # Grab the API 1000 packets
        # noinspection PyTypeChecker
        for packet in self.files_index.stream(
            io.ReadFilter.empty().with_api_versions({io.ApiVersion.API_1000})
        ):
            # noinspection Mypy
            result[packet.get_station_information().get_id()].append(packet)

        return result

    # noinspection PyTypeChecker
    def read_files_by_id(self, get_id: str) -> Optional[List[WrappedRedvoxPacketM]]:
        """
        :param get_id: the id to filter on
        :return: the list of packets with the requested id, or None if the id can't be found
        """

        result: List[WrappedRedvoxPacketM] = []

        # Iterate over the API 900 packets in a memory efficient way
        # and convert to API 1000
        for packet_900 in self.files_index.stream(
            io.ReadFilter.empty()
            .with_api_versions({io.ApiVersion.API_900})
            .with_station_ids({get_id})
        ):
            # noinspection Mypy
            result.append(ac.convert_api_900_to_1000(packet_900))

        # Grab the API 1000 packets
        for packet in self.files_index.stream(
            io.ReadFilter.empty()
            .with_api_versions({io.ApiVersion.API_1000})
            .with_station_ids({get_id})
        ):
            # noinspection Mypy
            result.append(packet)

        if len(result) == 0:
            return None

        return result

    def get_stations(self) -> List[Station]:
        """
        :return: a list of all stations represented by the data packets
        """
        station_ids: List[str] = self.index_summary.station_ids()
        pool = parallel.pool()

        stations_opt: List[Optional[Station]] = pool.map(self.get_station_by_id, station_ids)
        # noinspection Mypy
        return list(filter(lambda station: station is not None, stations_opt))

    def read_files_as_stations(self) -> Dict[str, Station]:
        """
        :return: a dictionary of all station_id to station pairs represented by the data packets
        """
        stations: List[Station] = self.get_stations()
        return {station.id: station for station in stations}

    def get_station_by_id(self, get_id: str) -> Optional[Station]:
        """
        :param get_id: the id to filter on
        :return: a station containing the data of the packets with the requested id or None if id can't be found
        """
        return Station(self.read_files_by_id(get_id))



import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression
from typing import Tuple


class OffsetModel:
    def __init__(self, latencies: np.ndarray, offsets: np.ndarray, times: np.ndarray,
                 k_bins: int, n_samples: int, start_time: float, end_time: float):
        self.slope, self.intercept = get_offset_function(latencies, offsets, times, k_bins,
                                                         n_samples, start_time, end_time)
        self.start_time = start_time
        self.end_time = end_time
        self.k_bins = k_bins
        self.n_samples = n_samples
        if all(np.nan_to_num(latencies) == 0.0):
            self.best_latency = np.nan
            self.best_offset = 0.0
        else:
            self.best_latency = np.nanmin(latencies[np.nonzero(latencies)])
            self.best_offset = offsets[np.argwhere(latencies == self.best_latency)[0][0]]

    def get_offset_at_new_time(self, new_time: float) -> float:
        """
        Get's offset at new_time time based on the offset model.
        :param new_time: The time of corresponding to the new offset
        :return: new offset corresponding to the new_time
        """
        return get_offset_at_new_time(new_time, self.slope, self.intercept, self.start_time)

    def update_time(self, new_time: float) -> float:
        """
        update new_time time based on the offset model.
        :param new_time: The time to update
        :return: updated new_time
        """
        return new_time + self.get_offset_at_new_time(new_time)


# The Weighted Linear Regression Function for offsets
def offset_weighted_linear_regression(latencies: np.ndarray, offsets: np.ndarray,
                                      times: np.ndarray) -> Tuple[float, float]:
    """
    Computes and returns the slope and intercept for the offset function (offset = slope * time + intercept)
    The intercept is based on first UTC time 0, all units are in microseconds
    :param latencies: array of the best latencies per packet
    :param offsets: array of offsets corresponding to the best latencies per packet
    :param times: array of device times corresponding to the best latencies per packet
    :return:  slope, intercept
    """

    if all(np.isnan(latencies)):
        return 0, 0

    # Compute the weights for the linear regression by the latencies
    base_weight = latencies.min() - 0.2
    deviations = latencies - base_weight
    weights = deviations ** -2

    # Set up the weighted linear regression
    wls = LinearRegression()
    wls.fit(X=times.reshape(-1, 1),
            y=offsets.reshape(-1, 1),
            sample_weight=weights)

    # return the slope and intercept
    return wls.coef_[0][0], wls.intercept_[0]


# Function to correct the intercept value
def get_offset_at_new_time(new_time: float, slope: float, intercept: float, model_time: float) -> float:
    """
    Get's offset at new_time time based on the offset model.
    :param new_time: The time of corresponding to the new offset
    :param slope: slope of the offset model
    :param intercept: the intercept of the offset model relative to the model_time
    :param model_time: the time corresponding to the intercept of the offset model
    :return: new offset corresponding to the new_time
    """
    # get the time difference
    time_diff = new_time - model_time

    # apply the offset model to get new intercept
    new_offset = slope*time_diff + intercept

    return new_offset


# Function to get the subset data frame to do the weighted linear regression
def get_binned_df(full_df: pd.DataFrame, bin_times: np.ndarray, n_samples: float) -> pd.DataFrame:
    """
    Returns a subset of the full_df with n_samples per binned times.
    nan latencies values will be ignored.
    :param full_df: pandas DataFrame containing latencies, offsets, and times.
    :param bin_times: array of edge times for each bin
    :param n_samples: number of samples to take per bin
    :return: binned_df
    """

    # Initialize the data frame
    binned_df = pd.DataFrame()

    # Loop through each bin and get the n smallest samples
    for i in range(len(bin_times)-1):
        # select the time range
        select_df = full_df[full_df['times'] < bin_times[i+1]]
        select_df = select_df[select_df['times'] > bin_times[i]]

        # select n_samples smallest values (ignores nan values)
        n_smallest = select_df.nsmallest(n_samples, 'latencies')

        # append the n_smallest entries
        binned_df = binned_df.append(n_smallest)

    # Sort the binned_df by time
    binned_df = binned_df.sort_values(by=['times'])

    return binned_df


# The main function
def get_offset_function(latencies: np.ndarray, offsets: np.ndarray, times: np.ndarray,
                        k_bins: int, n_samples: int, start_time: float, end_time: float) -> Tuple[float, float]:
    """
    Computes and returns the slope and intercept for the offset function (offset = slope * time + intercept)
    The data is binned by k_bins in equally spaced times, the n_samples best latencies are taken to be used to get
    the weighted linear regression. The slope and intercept is then returned as the offset model.
    The intercept is offset at first_start_time.
    :param latencies: array of the best latencies per packet
    :param offsets: array of offsets corresponding to the best latencies per packet
    :param times: array of device times corresponding to the best latencies per packet
    :param k_bins: number of bins to separate the latencies
    :param n_samples: number of points to use per bins
    :param start_time: the time used to compute the intercept (offset) and time bins; use start time of first packet
    :param end_time: the time used to compute the time bins; use start time of last packet + packet duration
    :return: slope, intercept
    """

    # Organize the data into a data frame
    full_df = pd.DataFrame(data=times, columns=['times'])
    full_df['latencies'] = latencies
    full_df['offsets'] = offsets

    # Get the index for the separations (add +1 to k_bins so that there would be k_bins bins)
    bin_times = np.linspace(start_time, end_time, k_bins + 1)

    # Make the dataframe with the data with n_samples per bins
    binned_df = get_binned_df(full_df=full_df,
                              bin_times=bin_times,
                              n_samples=n_samples)

    # Compute the weighted linear regression
    slope, zero_intercept = offset_weighted_linear_regression(latencies=binned_df['latencies'].values,
                                                              offsets=binned_df['offsets'].values,
                                                              times=binned_df['times'].values)

    # Get offset relative to the first time
    intercept = get_offset_at_new_time(new_time=start_time,
                                       slope=slope,
                                       intercept=zero_intercept,
                                       model_time=0)

    return slope, intercept



"""
Defines generic sensor data and data for API-independent analysis
all timestamps are integers in microseconds unless otherwise stated
"""
import enum
from typing import List, Optional, Tuple

import numpy as np
import pandas as pd

import redvox.common.date_time_utils as dtu
from redvox.common import offset_model as om
from redvox.api1000.wrapped_redvox_packet.wrapped_packet import WrappedRedvoxPacketM
from redvox.api1000.wrapped_redvox_packet.sensors import xyz, single


# todo: add original timestamps to the dataframes
class SensorType(enum.Enum):
    """
    Enumeration of possible types of sensors to read data from
    """

    UNKNOWN_SENSOR = 0  # unknown sensor
    ACCELEROMETER = 1  # meters/second^2
    AMBIENT_TEMPERATURE = 2  # degrees Celsius
    AUDIO = 3  # normalized counts
    COMPRESSED_AUDIO = 4  # bytes (codec specific)
    GRAVITY = 5  # meters/second^2
    GYROSCOPE = 6  # radians/second
    IMAGE = 7  # bytes (codec specific)
    LIGHT = 8  # lux
    LINEAR_ACCELERATION = 9  # meters/second^2
    LOCATION = 10  # See standard
    MAGNETOMETER = 11  # microtesla
    ORIENTATION = 12  # radians
    PRESSURE = 13  # kilopascal
    PROXIMITY = 14  # on, off, cm
    RELATIVE_HUMIDITY = 15  # percentage
    ROTATION_VECTOR = 16  # Unitless
    INFRARED = 17  # this is proximity
    STATION_HEALTH = 18
    # battery charge and current level, phone internal temperature, network source and strength,
    # available RAM of the system, cell service status, amount of hard disk space left, power charging state

    @staticmethod
    def type_from_str(type_str: str) -> "SensorType":
        """
        converts a string to a sensor type
        :param type_str: string to convert
        :return: a sensor type, UNKNOWN_SENSOR is the default for invalid inputs
        """
        if type_str.lower() == "mic" or type_str.lower() == "audio":
            return SensorType.AUDIO
        elif type_str.lower() == "accelerometer" or type_str.lower() == "accel":
            return SensorType.ACCELEROMETER
        elif type_str.lower() == "ambient_temperature":
            return SensorType.AMBIENT_TEMPERATURE
        elif type_str.lower() == "compressed_audio":
            return SensorType.COMPRESSED_AUDIO
        elif type_str.lower() == "gravity":
            return SensorType.GRAVITY
        elif type_str.lower() == "gyroscope" or type_str.lower() == "gyro":
            return SensorType.GYROSCOPE
        elif type_str.lower() == "image":
            return SensorType.IMAGE
        elif type_str.lower() == "light":
            return SensorType.LIGHT
        elif (
            type_str.lower() == "linear_acceleration"
            or type_str.lower() == "linear_accel"
        ):
            return SensorType.LINEAR_ACCELERATION
        elif type_str.lower() == "location" or type_str.lower() == "loc":
            return SensorType.LOCATION
        elif type_str.lower() == "magnetometer" or type_str.lower() == "mag":
            return SensorType.MAGNETOMETER
        elif type_str.lower() == "orientation":
            return SensorType.ORIENTATION
        elif (
            type_str.lower() == "pressure"
            or type_str.lower() == "bar"
            or type_str.lower() == "barometer"
        ):
            return SensorType.PRESSURE
        elif type_str.lower() == "proximity" or type_str.lower() == "infrared":
            return SensorType.PROXIMITY
        elif type_str.lower() == "relative_humidity":
            return SensorType.RELATIVE_HUMIDITY
        elif type_str.lower() == "rotation_vector":
            return SensorType.ROTATION_VECTOR
        else:
            return SensorType.UNKNOWN_SENSOR


class SensorData:
    """
    Generic SensorData class for API-independent analysis
    Properties:
        name: string, name of sensor
        type: SensorType, enumerated type of sensor
        data_df: dataframe of the sensor data; always has timestamps as the first column,
                    the other columns are the data fields
        sample_rate: float, sample rate in Hz of the sensor, default np.nan, usually 1/sample_interval_s
        sample_interval_s: float, mean duration in seconds between samples, default np.nan, usually 1/sample_rate
        sample_interval_std_s: float, standard deviation in seconds between samples, default np.nan
        is_sample_rate_fixed: bool, True if sample rate is constant, default False
        timestamps_altered: bool, True if timestamps in the sensor have been altered from their original values
                            default False
    """

    def __init__(
        self,
        sensor_name: str,
        sensor_data: pd.DataFrame,
        sensor_type: SensorType = SensorType.UNKNOWN_SENSOR,
        sample_rate: float = np.nan,
        sample_interval_s: float = np.nan,
        sample_interval_std_s: float = np.nan,
        is_sample_rate_fixed: bool = False,
        are_timestamps_altered: bool = False,
        calculate_stats: bool = False
    ):
        """
        initialize the sensor data with params
        :param sensor_name: name of the sensor
        :param sensor_type: enumerated type of the sensor, default SensorType.UNKNOWN_SENSOR
        :param sensor_data: dataframe with the timestamps and sensor data; first column is always the timestamps,
                            the other columns are the data channels in the sensor
        :param sample_rate: sample rate in hz of the data
        :param sample_interval_s: sample interval in seconds of the data
        :param sample_interval_std_s: std dev of sample interval in seconds of the data
        :param is_sample_rate_fixed: if True, sample rate is constant for all data, default False
        :param are_timestamps_altered: if True, timestamps in the sensor have been altered from their
                                        original values, default False
        :param calculate_stats: if True, calculate sample_rate, sample_interval_s, and sample_interval_std_s
                                default False
        """
        if "timestamps" not in sensor_data.columns:
            raise AttributeError(
                'SensorData requires the data frame to contain a column titled "timestamps"'
            )
        self.name: str = sensor_name
        self.type: SensorType = sensor_type
        self.data_df: pd.DataFrame = sensor_data.infer_objects()
        self.sample_rate: float = sample_rate
        self.sample_interval_s: float = sample_interval_s
        self.sample_interval_std_s: float = sample_interval_std_s
        self.is_sample_rate_fixed: bool = is_sample_rate_fixed
        self.timestamps_altered: bool = are_timestamps_altered
        if calculate_stats:
            self.organize_and_update_stats()
        else:
            self.sort_by_data_timestamps()
        # todo: store the non-consecutive timestamp indices (idk how to find those)

    def is_sample_interval_invalid(self) -> bool:
        """
        :return: True if sample interval is np.nan or equal to 0.0
        """
        return np.isnan(self.sample_interval_s) or self.sample_interval_s == 0.0

    def organize_and_update_stats(self) -> "SensorData":
        """
        sorts the data by timestamps, then if the sample rate is not fixed, recalculates the sample rate, interval,
            and interval std dev.  If there is only one value, sets the sample rate, interval, and interval std dev
            to np.nan.  Updates the SensorData object with the new values
        :return: updated version of self
        """
        self.sort_by_data_timestamps()
        if not self.is_sample_rate_fixed:
            if self.num_samples() > 1:
                timestamp_diffs = np.diff(self.data_timestamps())
                self.sample_interval_s = dtu.microseconds_to_seconds(
                    float(np.mean(timestamp_diffs))
                )
                self.sample_interval_std_s = dtu.microseconds_to_seconds(
                    float(np.std(timestamp_diffs))
                )
                self.sample_rate = (
                    np.nan
                    if self.is_sample_interval_invalid()
                    else 1 / self.sample_interval_s
                )
            else:
                self.sample_interval_s = np.nan
                self.sample_interval_std_s = np.nan
                self.sample_rate = np.nan
        return self

    def append_data(
        self, new_data: pd.DataFrame, recalculate_stats: bool = False
    ) -> "SensorData":
        """
        append the new data to the dataframe, update the sensor's stats on demand if it doesn't have a fixed
            sample rate, then return the updated SensorData object
        :param new_data: Dataframe containing data to add to the sensor's dataframe
        :param recalculate_stats: if True and the sensor does not have a fixed sample rate, sort the timestamps,
                                    recalculate the sample rate, interval, and interval std dev, default False
        :return: the updated SensorData object
        """
        self.data_df = self.data_df.append(new_data, ignore_index=True)
        if recalculate_stats and not self.is_sample_rate_fixed:
            self.organize_and_update_stats()
        return self

    def sensor_type_as_str(self) -> str:
        """
        gets the sensor type as a string
        :return: sensor type of the sensor as a string
        """
        return self.type.name

    def samples(self) -> np.ndarray:
        """
        gets the samples of dataframe
        :return: the data values of the dataframe as a numpy ndarray
        """
        return self.data_df.iloc[:, 1:].T.to_numpy()

    def get_data_channel(self, channel_name: str) -> np.array:
        """
        gets the data channel specified, raises an error and lists valid fields if channel_name is not in the dataframe
        :param channel_name: the name of the channel to get data for
        :return: the data values of the channel as a numpy array or a list of strings if the channel is enumerated
        """
        if channel_name not in self.data_df.columns:
            raise ValueError(
                f"WARNING: {channel_name} does not exist; try one of {self.data_channels()}"
            )
        return self.data_df[channel_name].to_numpy()

    def get_valid_data_channel_values(self, channel_name: str) -> np.array:
        """
        gets all non-nan values from the channel specified
        :param channel_name: the name of the channel to get data for
        :return: non-nan values of the channel as a numpy array
        """
        channel_data = self.get_data_channel(channel_name)
        return channel_data[~np.isnan(channel_data)]

    def data_timestamps(self) -> np.array:
        """
        :return: the timestamps as a numpy array
        """
        return self.data_df["timestamps"].to_numpy(dtype=np.float)

    def first_data_timestamp(self) -> float:
        """
        :return: timestamp of the first data point
        """
        return self.data_df["timestamps"].iloc[0]

    def last_data_timestamp(self) -> float:
        """
        :return: timestamp of the last data point
        """
        return self.data_df["timestamps"].iloc[-1]

    def num_samples(self) -> int:
        """
        :return: the number of rows (samples) in the dataframe
        """
        return self.data_df.shape[0]

    def data_channels(self) -> List[str]:
        """
        :return: a list of the names of the columns (data channels) of the dataframe
        """
        return self.data_df.columns.to_list()

    def update_data_timestamps(self, offset_model: om.OffsetModel):
        self.data_df["timestamps"] = [offset_model.start_time + offset_model.intercept + t * (1 + offset_model.slope)
                                      for t in self.data_timestamps()]
        time_diffs = np.floor(np.diff(self.data_timestamps()))
        if len(time_diffs) > 1:
            self.sample_interval_s += offset_model.slope
            if self.sample_interval_s > 0:
                self.sample_rate = 1 / self.sample_interval_s
                self.sample_interval_std_s = dtu.microseconds_to_seconds(np.std(time_diffs))
            self.timestamps_altered = True

    def update_data_timestamps_delta(self, time_delta: float):
        """
        adds the time_delta to the sensor's timestamps; use negative values to go backwards in time
        :param time_delta: time to add to sensor's timestamps
        """
        new_timestamps = self.data_timestamps() + time_delta
        self.data_df["timestamps"] = new_timestamps
        self.timestamps_altered = True

    def sort_by_data_timestamps(self, ascending: bool = True):
        """
        sorts the data based on timestamps
        :param ascending: if True, timestamps are sorted in ascending order
        """
        self.data_df = self.data_df.sort_values("timestamps", ascending=ascending)

    def interpolate(self, first_point: int, second_point: int, interpolate_timestamp: float) -> pd.Series:
        """
        interpolates two points at the intercept value
        :param first_point: index of first point
        :param second_point: index of second point
        :param interpolate_timestamp: timestamp to be interpolate other values
        :return: pd.Series of interpolated points
        """
        numeric_series = self.data_df.select_dtypes(include=[np.number])
        numeric_diff = numeric_series.iloc[second_point] - numeric_series.iloc[first_point]
        numeric_diff = (numeric_diff / numeric_diff["timestamps"]) \
            * (interpolate_timestamp - numeric_series["timestamps"][first_point]) \
            + numeric_series.iloc[first_point]
        non_numeric_series = self.data_df.select_dtypes(exclude=[np.number])
        if np.abs(self.data_df.iloc[first_point]["timestamps"] - interpolate_timestamp) \
                <= np.abs(self.data_df.iloc[second_point]["timestamps"] - interpolate_timestamp):
            non_numeric_diff = non_numeric_series.iloc[first_point]
        else:
            non_numeric_diff = non_numeric_series.iloc[second_point]
        return pd.concat([numeric_diff, non_numeric_diff])


def get_empty_sensor_data(
    name: str, sensor_type: SensorType = SensorType.UNKNOWN_SENSOR
) -> SensorData:
    """
    create a sensor data object with no data
    :param name: name of the sensor
    :param sensor_type: type of the sensor to create, default SensorType.UNKNOWN_SENSOR
    :return: empty sensor
    """
    return SensorData(name, pd.DataFrame([], columns=["timestamps"]), sensor_type)


def calc_evenly_sampled_timestamps(
    start: float, samples: int, rate_hz: float
) -> np.array:
    """
    given a start time, calculates samples amount of evenly spaced timestamps at rate_hz
    :param start: float, start timestamp in microseconds
    :param samples: int, number of samples
    :param rate_hz: float, sample rate in hz
    :return: np.array with evenly spaced timestamps starting at start
    """
    return start + (np.arange(0, samples) / rate_hz) * dtu.MICROSECONDS_IN_SECOND


def get_sample_statistics(data_df: pd.DataFrame) -> Tuple[float, float, float]:
    """
    calculate the sample rate, interval and interval std dev using the timestamps in the dataframe
    :param data_df: the dataframe containing timestamps to calculate statistics from
    :return: a Tuple containing the sample rate, interval and interval std dev
    """
    if data_df["timestamps"].size > 1:
        sample_interval = dtu.microseconds_to_seconds(
            float(np.mean(np.diff(data_df["timestamps"])))
        )
        sample_interval_std = dtu.microseconds_to_seconds(
            float(np.std(np.diff(data_df["timestamps"])))
        )
    else:
        sample_interval = np.nan
        sample_interval_std = np.nan
    return 1 / sample_interval, sample_interval, sample_interval_std


def read_apim_xyz_sensor(sensor: xyz.Xyz, column_id: str) -> pd.DataFrame:
    """
    read a sensor that has xyz data channels from an api M data packet
    raises Attribute Error if sensor does not contain xyz channels
    :param sensor: the xyz api M sensor to read
    :param column_id: string, used to name the columns
    :return: Dataframe representing the data in the sensor
    """
    timestamps = sensor.get_timestamps().get_timestamps()
    try:
        columns = ["timestamps", f"{column_id}_x", f"{column_id}_y", f"{column_id}_z"]
        return pd.DataFrame(
            np.transpose(
                [
                    timestamps,
                    sensor.get_x_samples().get_values(),
                    sensor.get_y_samples().get_values(),
                    sensor.get_z_samples().get_values(),
                ]
            ),
            columns=columns,
        )
    except AttributeError:
        raise


def load_apim_xyz_sensor(sensor_type: SensorType, data, name, description) -> Optional[SensorData]:
    if len(data[0]) > 0:
        return SensorData(
            description,
            pd.DataFrame(np.transpose([data[0], data[1], data[2], data[3]]),
                         columns=["timestamps", f"{name}_x", f"{name}_y", f"{name}_z"]),
            sensor_type,
            calculate_stats=True
        )
    return None


def read_apim_single_sensor(sensor: single.Single, column_id: str) -> pd.DataFrame:
    """
    read a sensor that has a single data channel from an api M data packet
    raises Attribute Error if sensor does not contain exactly one data channel
    :param sensor: the single channel api M sensor to read
    :param column_id: string, used to name the columns
    :return: Dataframe representing the data in the sensor
    """
    timestamps = sensor.get_timestamps().get_timestamps()
    try:
        columns = ["timestamps", column_id]
        return pd.DataFrame(
            np.transpose([timestamps, sensor.get_samples().get_values()]),
            columns=columns,
        )
    except AttributeError:
        raise


def load_apim_audio(wrapped_packet: WrappedRedvoxPacketM) -> Optional[SensorData]:
    """
    load audio data from a single wrapped packet
    :param wrapped_packet: packet with data to load
    :return: audio sensor data if it exists, None otherwise
    """
    audio = wrapped_packet.get_sensors().get_audio()
    if audio and wrapped_packet.get_sensors().validate_audio():
        sample_rate_hz = audio.get_sample_rate()
        data_for_df = audio.get_samples().get_values()
        timestamps = calc_evenly_sampled_timestamps(
            audio.get_first_sample_timestamp(), audio.get_num_samples(), sample_rate_hz
        )
        return SensorData(
            audio.get_sensor_description(),
            pd.DataFrame(
                np.transpose([timestamps, data_for_df]),
                columns=["timestamps", "microphone"],
            ),
            SensorType.AUDIO,
            sample_rate_hz,
            1 / sample_rate_hz,
            0.0,
            True,
        )
    return None


def load_apim_audio_from_list(wrapped_packets: List[WrappedRedvoxPacketM]) -> Optional[SensorData]:
    """
    load audio data from a list of wrapped packets
    NOTE: This only works because audio sensors in the list should all have the same number of data points.
    :param wrapped_packets: packets with data to load
    :return: audio sensor data if it exists, None otherwise
    """
    if len(wrapped_packets) > 0:
        if wrapped_packets[0].get_sensors().get_audio() and wrapped_packets[0].get_sensors().validate_audio():
            try:
                sample_rate_hz = wrapped_packets[0].get_sensors().get_audio().get_sample_rate()
                timestamps = np.array(
                    [calc_evenly_sampled_timestamps(p.get_sensors().get_audio().get_first_sample_timestamp(),
                                                    p.get_sensors().get_audio().get_num_samples(), sample_rate_hz)
                     for p in wrapped_packets]).flatten()
                data_vals = np.array([p.get_sensors().get_audio().get_samples().get_values()
                                      for p in wrapped_packets]).flatten()
                return SensorData(
                    wrapped_packets[0].get_sensors().get_audio().get_sensor_description(),
                    pd.DataFrame(np.transpose([timestamps, data_vals]), columns=["timestamps", "microphone"]),
                    SensorType.AUDIO,
                    sample_rate_hz,
                    1 / sample_rate_hz,
                    0.0,
                    True,
                    )
            except ValueError:
                print("Data arrays do not have the same number of points.\n"
                      "Original error message: ", ValueError)
    return None


def load_apim_compressed_audio(wrapped_packet: WrappedRedvoxPacketM) -> Optional[SensorData]:
    """
    load compressed audio data from a single wrapped packet
    :param wrapped_packet: packet with data to load
    :return: compressed audio sensor data if it exists, None otherwise
    """
    comp_audio = wrapped_packet.get_sensors().get_compressed_audio()
    if comp_audio and wrapped_packet.get_sensors().validate_compressed_audio():
        sample_rate_hz = comp_audio.get_sample_rate()
        return SensorData(
            comp_audio.get_sensor_description(),
            pd.DataFrame(
                np.transpose(
                    [
                        comp_audio.get_first_sample_timestamp(),
                        comp_audio.get_audio_bytes(),
                        comp_audio.get_audio_codec(),
                    ]
                ),
                columns=["timestamps", "compressed_audio", "audio_codec"],
            ),
            SensorType.COMPRESSED_AUDIO,
            sample_rate_hz,
            1 / sample_rate_hz,
            0.0,
            True,
        )
    return None


def load_apim_compressed_audio_from_list(wrapped_packets: List[WrappedRedvoxPacketM]) -> Optional[SensorData]:
    """
    load compressed audio data from a list of wrapped packets
    :param wrapped_packets: packets with data to load
    :return: compressed audio sensor data if it exists, None otherwise
    """
    data_df = [[], [], []]
    for packet in wrapped_packets:
        comp_audio = packet.get_sensors().get_compressed_audio()
        if comp_audio and packet.get_sensors().validate_compressed_audio():
            data_df[0].append(comp_audio.get_first_sample_timestamp())
            data_df[1].append(comp_audio.get_audio_bytes())
            data_df[2].append(comp_audio.get_audio_codec())
    if len(data_df[0]) > 0:
        sample_rate_hz = wrapped_packets[0].get_sensors().get_compressed_audio().get_sample_rate()
        return SensorData(
            wrapped_packets[0].get_sensors().get_compressed_audio().get_sensor_description(),
            pd.DataFrame(
                np.transpose([data_df[0], data_df[1], data_df[2]]),
                columns=["timestamps", "compressed_audio", "audio_codec"],
            ),
            SensorType.COMPRESSED_AUDIO,
            sample_rate_hz,
            1 / sample_rate_hz,
            0.0,
            True,
            )
    return None


def load_apim_image(wrapped_packet: WrappedRedvoxPacketM) -> Optional[SensorData]:
    """
    load image data from a single wrapped packet
    :param wrapped_packet: packet with data to load
    :return: image sensor data if it exists, None otherwise
    """
    image = wrapped_packet.get_sensors().get_image()
    if image and wrapped_packet.get_sensors().validate_image():
        timestamps = image.get_timestamps().get_timestamps()
        codecs = np.full(len(timestamps), image.get_image_codec().value)
        data_df = pd.DataFrame(
            np.transpose([timestamps, image.get_samples(), codecs]),
            columns=["timestamps", "image", "image_codec"],
        )
        sample_rate, sample_interval, sample_interval_std = get_sample_statistics(data_df)
        return SensorData(
            image.get_sensor_description(),
            data_df,
            SensorType.IMAGE,
            sample_rate,
            sample_interval,
            sample_interval_std,
            False,
        )
    return None


def load_apim_image_from_list(wrapped_packets: List[WrappedRedvoxPacketM]) -> Optional[SensorData]:
    """
    load image data from a list of wrapped packets
    :param wrapped_packets: packets with data to load
    :return: image sensor data if it exists, None otherwise
    """
    data_list = [[], [], []]
    for packet in wrapped_packets:
        image = packet.get_sensors().get_image()
        if image and packet.get_sensors().validate_image():
            data_list[0] = image.get_timestamps().get_timestamps()
            data_list[1] = image.get_samples()
            data_list[2] = np.full(len(data_list[0]), image.get_image_codec().value)
    if len(data_list[0]) > 0:
        data_df = pd.DataFrame(
            np.transpose([data_list[0], data_list[1], data_list[2]]),
            columns=["timestamps", "image", "image_codec"],
        )
        sample_rate, sample_interval, sample_interval_std = get_sample_statistics(data_df)
        return SensorData(
            wrapped_packets[0].get_sensors().get_image().get_sensor_description(),
            data_df,
            SensorType.IMAGE,
            sample_rate,
            sample_interval,
            sample_interval_std,
            False,
        )
    return None


def load_apim_location(wrapped_packet: WrappedRedvoxPacketM) -> Optional[SensorData]:
    """
    load location data from a single wrapped packet
    :param wrapped_packet: packet with data to load
    :return: location sensor data if it exists, None otherwise
    """
    loc = wrapped_packet.get_sensors().get_location()
    if loc and wrapped_packet.get_sensors().validate_location():
        if loc.is_only_best_values():
            if loc.get_last_best_location():
                best_loc = loc.get_last_best_location()
            else:
                best_loc = loc.get_overall_best_location()
            data_for_df = [
                [
                    best_loc.get_latitude_longitude_timestamp().get_mach(),
                    best_loc.get_latitude(),
                    best_loc.get_longitude(),
                    best_loc.get_altitude(),
                    best_loc.get_speed(),
                    best_loc.get_bearing(),
                    best_loc.get_horizontal_accuracy(),
                    best_loc.get_vertical_accuracy(),
                    best_loc.get_speed_accuracy(),
                    best_loc.get_bearing_accuracy(),
                    best_loc.get_location_provider(),
                ]
            ]
        else:
            timestamps = loc.get_timestamps().get_timestamps()
            if len(timestamps) > 0:
                lat_samples = loc.get_latitude_samples().get_values()
                lon_samples = loc.get_longitude_samples().get_values()
                alt_samples = loc.get_altitude_samples().get_values()
                spd_samples = loc.get_speed_samples().get_values()
                bear_samples = loc.get_bearing_samples().get_values()
                hor_acc_samples = loc.get_horizontal_accuracy_samples().get_values()
                vert_acc_samples = loc.get_vertical_accuracy_samples().get_values()
                spd_acc_samples = loc.get_speed_accuracy_samples().get_values()
                bear_acc_samples = loc.get_bearing_accuracy_samples().get_values()
                loc_prov_samples = loc.get_location_providers().get_values()
                data_for_df = []
                for i in range(len(timestamps)):
                    new_entry = [
                        timestamps[i],
                        lat_samples[i],
                        lon_samples[i],
                        np.nan if len(alt_samples) < i + 1 else alt_samples[i],
                        np.nan if len(spd_samples) < i + 1 else spd_samples[i],
                        np.nan if len(bear_samples) < i + 1 else bear_samples[i],
                        np.nan if len(hor_acc_samples) < i + 1 else hor_acc_samples[i],
                        np.nan if len(vert_acc_samples) < i + 1 else vert_acc_samples[i],
                        np.nan if len(spd_acc_samples) < i + 1 else spd_acc_samples[i],
                        np.nan if len(bear_acc_samples) < i + 1 else bear_acc_samples[i],
                        np.nan if len(loc_prov_samples) < i + 1 else loc_prov_samples[i],
                    ]
                    data_for_df.append(new_entry)
            else:
                return None
        data_df = pd.DataFrame(
            data_for_df,
            columns=[
                "timestamps",
                "latitude",
                "longitude",
                "altitude",
                "speed",
                "bearing",
                "horizontal_accuracy",
                "vertical_accuracy",
                "speed_accuracy",
                "bearing_accuracy",
                "location_provider",
            ],
        )
        sample_rate, sample_interval, sample_interval_std = get_sample_statistics(data_df)
        return SensorData(
            loc.get_sensor_description(),
            data_df,
            SensorType.LOCATION,
            sample_rate,
            sample_interval,
            sample_interval_std,
            False,
        )
    return None


def load_apim_location_from_list(wrapped_packets: List[WrappedRedvoxPacketM]) -> Optional[SensorData]:
    """
    load location data from a list of wrapped packets
    :param wrapped_packets: packets with data to load
    :return: location sensor data if it exists, None otherwise
    """
    data_for_df = [[], [], [], [], [], [], [], [], [], [], []]
    for packet in wrapped_packets:
        loc = packet.get_sensors().get_location()
        if loc and packet.get_sensors().validate_location():
            if loc.is_only_best_values():
                if loc.get_last_best_location():
                    best_loc = loc.get_last_best_location()
                else:
                    best_loc = loc.get_overall_best_location()
                data_for_df[0].append(best_loc.get_latitude_longitude_timestamp().get_mach())
                data_for_df[1].append(best_loc.get_latitude())
                data_for_df[2].append(best_loc.get_longitude())
                data_for_df[3].append(best_loc.get_altitude())
                data_for_df[4].append(best_loc.get_speed())
                data_for_df[5].append(best_loc.get_bearing())
                data_for_df[6].append(best_loc.get_horizontal_accuracy())
                data_for_df[7].append(best_loc.get_vertical_accuracy())
                data_for_df[8].append(best_loc.get_speed_accuracy())
                data_for_df[9].append(best_loc.get_bearing_accuracy())
                data_for_df[10].append(best_loc.get_location_provider())
            else:
                timestamps = loc.get_timestamps().get_timestamps()
                if len(timestamps) > 0:
                    data_for_df[0].extend(timestamps)
                    data_for_df[1].extend(loc.get_latitude_samples().get_values())
                    data_for_df[2].extend(loc.get_longitude_samples().get_values())
                    alt_samples = loc.get_altitude_samples().get_values()
                    if len(alt_samples) <= 0:
                        alt_samples = np.full(len(timestamps), np.nan)
                    data_for_df[3].extend(alt_samples)
                    spd_samples = loc.get_speed_samples().get_values()
                    if len(spd_samples) <= 0:
                        spd_samples = np.full(len(timestamps), np.nan)
                    data_for_df[4].extend(spd_samples)
                    bear_samples = loc.get_bearing_samples().get_values()
                    if len(bear_samples) <= 0:
                        bear_samples = np.full(len(timestamps), np.nan)
                    data_for_df[5].extend(bear_samples)
                    hor_acc_samples = loc.get_horizontal_accuracy_samples().get_values()
                    if len(hor_acc_samples) <= 0:
                        hor_acc_samples = np.full(len(timestamps), np.nan)
                    data_for_df[6].extend(hor_acc_samples)
                    vert_acc_samples = loc.get_vertical_accuracy_samples().get_values()
                    if len(vert_acc_samples) <= 0:
                        vert_acc_samples = np.full(len(timestamps), np.nan)
                    data_for_df[7].extend(vert_acc_samples)
                    spd_acc_samples = loc.get_speed_accuracy_samples().get_values()
                    if len(spd_acc_samples) <= 0:
                        spd_acc_samples = np.full(len(timestamps), np.nan)
                    data_for_df[8].extend(spd_acc_samples)
                    bear_acc_samples = loc.get_bearing_accuracy_samples().get_values()
                    if len(bear_acc_samples) <= 0:
                        bear_acc_samples = np.full(len(timestamps), np.nan)
                    data_for_df[9].extend(bear_acc_samples)
                    loc_prov_samples = loc.get_location_providers().get_values()
                    if len(loc_prov_samples) <= len(timestamps):
                        loc_prov_samples = np.full(len(timestamps), loc_prov_samples[0])
                    data_for_df[10].extend(loc_prov_samples)
    if len(data_for_df[0]) > 0:
        return SensorData(
            wrapped_packets[0].get_sensors().get_location().get_sensor_description(),
            pd.DataFrame(np.transpose(data_for_df), columns=[
                "timestamps",
                "latitude",
                "longitude",
                "altitude",
                "speed",
                "bearing",
                "horizontal_accuracy",
                "vertical_accuracy",
                "speed_accuracy",
                "bearing_accuracy",
                "location_provider",
            ]),
            SensorType.LOCATION,
            calculate_stats=True
        )
    return None


def load_apim_pressure(wrapped_packet: WrappedRedvoxPacketM) -> Optional[SensorData]:
    """
    load pressure data from a single wrapped packet
    :param wrapped_packet: packet with data to load
    :return: pressure sensor data if it exists, None otherwise
    """
    pressure = wrapped_packet.get_sensors().get_pressure()
    if pressure and wrapped_packet.get_sensors().validate_pressure():
        data_df = read_apim_single_sensor(pressure, "pressure")
        sample_rate, sample_interval, sample_interval_std = get_sample_statistics(data_df)
        return SensorData(
            pressure.get_sensor_description(),
            data_df,
            SensorType.PRESSURE,
            sample_rate,
            sample_interval,
            sample_interval_std,
            False,
        )


def load_apim_pressure_from_list(wrapped_packets: List[WrappedRedvoxPacketM]) -> Optional[SensorData]:
    """
    load pressure data from a list of wrapped packets
    :param wrapped_packets: packets with data to load
    :return: pressure sensor data if it exists, None otherwise
    """
    data_df = []
    timestamps = []
    for packet in wrapped_packets:
        pressure = packet.get_sensors().get_pressure()
        if pressure and packet.get_sensors().validate_pressure():
            data_df.extend(pressure.get_samples().get_values())
            timestamps.extend(pressure.get_timestamps().get_timestamps())
    if len(data_df) > 0:
        return SensorData(
            wrapped_packets[0].get_sensors().get_pressure().get_sensor_description(),
            pd.DataFrame(np.transpose([timestamps, data_df]), columns=["timestamps", "pressure"]),
            SensorType.PRESSURE,
            calculate_stats=True
        )
    return None


def load_apim_light(wrapped_packet: WrappedRedvoxPacketM) -> Optional[SensorData]:
    """
    load light data from a single wrapped packet
    :param wrapped_packet: packet with data to load
    :return: light sensor data if it exists, None otherwise
    """
    light = wrapped_packet.get_sensors().get_light()
    if light and wrapped_packet.get_sensors().validate_light():
        data_df = read_apim_single_sensor(light, "light")
        sample_rate, sample_interval, sample_interval_std = get_sample_statistics(data_df)
        return SensorData(
            light.get_sensor_description(),
            data_df,
            SensorType.LIGHT,
            sample_rate,
            sample_interval,
            sample_interval_std,
            False,
        )


def load_apim_light_from_list(wrapped_packets: List[WrappedRedvoxPacketM]) -> Optional[SensorData]:
    """
    load light data from a list of wrapped packets
    :param wrapped_packets: packets with data to load
    :return: light sensor data if it exists, None otherwise
    """
    data_df = []
    timestamps = []
    for packet in wrapped_packets:
        light = packet.get_sensors().get_light()
        if light and packet.get_sensors().validate_light():
            data_df.extend(light.get_samples().get_values())
            timestamps.extend(light.get_timestamps().get_timestamps())
    if len(data_df) > 0:
        return SensorData(
            wrapped_packets[0].get_sensors().get_light().get_sensor_description(),
            pd.DataFrame(np.transpose([timestamps, data_df]), columns=["timestamps", "light"]),
            SensorType.LIGHT,
            calculate_stats=True
        )
    return None


def load_apim_proximity(wrapped_packet: WrappedRedvoxPacketM) -> Optional[SensorData]:
    """
    load proximity data from a single wrapped packet
    :param wrapped_packet: packet with data to load
    :return: proximity sensor data if it exists, None otherwise
    """
    proximity = wrapped_packet.get_sensors().get_proximity()
    if proximity and wrapped_packet.get_sensors().validate_proximity():
        data_df = read_apim_single_sensor(proximity, "proximity")
        sample_rate, sample_interval, sample_interval_std = get_sample_statistics(data_df)
        return SensorData(
            proximity.get_sensor_description(),
            data_df,
            SensorType.PROXIMITY,
            sample_rate,
            sample_interval,
            sample_interval_std,
            False,
        )


def load_apim_proximity_from_list(wrapped_packets: List[WrappedRedvoxPacketM]) -> Optional[SensorData]:
    """
    load proximity data from a list of wrapped packets
    :param wrapped_packets: packets with data to load
    :return: proximity sensor data if it exists, None otherwise
    """
    data_df = []
    timestamps = []
    for packet in wrapped_packets:
        proximity = packet.get_sensors().get_proximity()
        if proximity and packet.get_sensors().validate_proximity():
            data_df.extend(proximity.get_samples().get_values())
            timestamps.extend(proximity.get_timestamps().get_timestamps())
    if len(data_df) > 0:
        return SensorData(
            wrapped_packets[0].get_sensors().get_proximity().get_sensor_description(),
            pd.DataFrame(np.transpose([timestamps, data_df]), columns=["timestamps", "proximity"]),
            SensorType.PROXIMITY,
            calculate_stats=True
        )
    return None


def load_apim_ambient_temp(wrapped_packet: WrappedRedvoxPacketM) -> Optional[SensorData]:
    """
    load ambient temperature data from a single wrapped packet
    :param wrapped_packet: packet with data to load
    :return: ambient temperature sensor data if it exists, None otherwise
    """
    ambient_temp = wrapped_packet.get_sensors().get_ambient_temperature()
    if ambient_temp and wrapped_packet.get_sensors().validate_ambient_temperature():
        data_df = read_apim_single_sensor(ambient_temp, "ambient_temp")
        sample_rate, sample_interval, sample_interval_std = get_sample_statistics(data_df)
        return SensorData(
            ambient_temp.get_sensor_description(),
            data_df,
            SensorType.AMBIENT_TEMPERATURE,
            sample_rate,
            sample_interval,
            sample_interval_std,
            False,
        )


def load_apim_ambient_temp_from_list(wrapped_packets: List[WrappedRedvoxPacketM]) -> Optional[SensorData]:
    """
    load ambient temperature data from a list of wrapped packets
    :param wrapped_packets: packets with data to load
    :return: ambient temperature sensor data if it exists, None otherwise
    """
    data_df = []
    timestamps = []
    for packet in wrapped_packets:
        amb_temp = packet.get_sensors().get_ambient_temperature()
        if amb_temp and packet.get_sensors().validate_ambient_temperature():
            data_df.extend(amb_temp.get_samples().get_values())
            timestamps.extend(amb_temp.get_timestamps().get_timestamps())
    if len(data_df) > 0:
        return SensorData(
            wrapped_packets[0].get_sensors().get_ambient_temperature().get_sensor_description(),
            pd.DataFrame(np.transpose([timestamps, data_df]), columns=["timestamps", "ambient_temp"]),
            SensorType.AMBIENT_TEMPERATURE,
            calculate_stats=True
        )
    return None


def load_apim_rel_humidity(wrapped_packet: WrappedRedvoxPacketM) -> Optional[SensorData]:
    """
    load relative humidity data from a single wrapped packet
    :param wrapped_packet: packet with data to load
    :return: relative humidity sensor data if it exists, None otherwise
    """
    rel_humidity = wrapped_packet.get_sensors().get_relative_humidity()
    if rel_humidity and wrapped_packet.get_sensors().validate_relative_humidity():
        data_df = read_apim_single_sensor(rel_humidity, "rel_humidity")
        sample_rate, sample_interval, sample_interval_std = get_sample_statistics(data_df)
        return SensorData(
            rel_humidity.get_sensor_description(),
            data_df,
            SensorType.RELATIVE_HUMIDITY,
            sample_rate,
            sample_interval,
            sample_interval_std,
            False,
        )


def load_apim_rel_humidity_from_list(wrapped_packets: List[WrappedRedvoxPacketM]) -> Optional[SensorData]:
    """
    load relative humidity data from a list of wrapped packets
    :param wrapped_packets: packets with data to load
    :return: relative humidity sensor data if it exists, None otherwise
    """
    data_df = []
    timestamps = []
    for packet in wrapped_packets:
        rel_hum = packet.get_sensors().get_relative_humidity()
        if rel_hum and packet.get_sensors().validate_relative_humidity():
            data_df.extend(rel_hum.get_samples().get_values())
            timestamps.extend(rel_hum.get_timestamps().get_timestamps())
    if len(data_df) > 0:
        return SensorData(
            wrapped_packets[0].get_sensors().get_relative_humidity().get_sensor_description(),
            pd.DataFrame(np.transpose([timestamps, data_df]), columns=["timestamps", "rel_humidity"]),
            SensorType.RELATIVE_HUMIDITY,
            calculate_stats=True
        )
    return None


def load_apim_accelerometer(wrapped_packet: WrappedRedvoxPacketM) -> Optional[SensorData]:
    """
    load accelerometer data from a single wrapped packet
    :param wrapped_packet: packet with data to load
    :return: accelerometer sensor data if it exists, None otherwise
    """
    accel = wrapped_packet.get_sensors().get_accelerometer()
    if accel and wrapped_packet.get_sensors().validate_accelerometer():
        data_df = read_apim_xyz_sensor(accel, "accelerometer")
        sample_rate, sample_interval, sample_interval_std = get_sample_statistics(data_df)
        return SensorData(
            accel.get_sensor_description(),
            data_df,
            SensorType.ACCELEROMETER,
            sample_rate,
            sample_interval,
            sample_interval_std,
            False,
        )


def load_apim_accelerometer_from_list(wrapped_packets: List[WrappedRedvoxPacketM]) -> Optional[SensorData]:
    """
    load accelerometer data from a list of wrapped packets
    :param wrapped_packets: packets with data to load
    :return: accelerometer sensor data if it exists, None otherwise
    """
    data_df = [[], [], [], []]
    for packet in wrapped_packets:
        accel = packet.get_sensors().get_accelerometer()
        if accel and packet.get_sensors().validate_accelerometer():
            data_df[0].extend(accel.get_timestamps().get_timestamps())
            data_df[1].extend(accel.get_x_samples().get_values())
            data_df[2].extend(accel.get_y_samples().get_values())
            data_df[3].extend(accel.get_z_samples().get_values())
    if len(data_df[0]) > 0:
        return load_apim_xyz_sensor(SensorType.ACCELEROMETER, data_df, "accelerometer",
                                    wrapped_packets[0].get_sensors().get_accelerometer().get_sensor_description())
    return None


def load_apim_magnetometer(wrapped_packet: WrappedRedvoxPacketM) -> Optional[SensorData]:
    """
    load magnetometer data from a single wrapped packet
    :param wrapped_packet: packet with data to load
    :return: magnetometer sensor data if it exists, None otherwise
    """
    mag = wrapped_packet.get_sensors().get_magnetometer()
    if mag and wrapped_packet.get_sensors().validate_magnetometer():
        data_df = read_apim_xyz_sensor(mag, "magnetometer")
        sample_rate, sample_interval, sample_interval_std = get_sample_statistics(data_df)
        return SensorData(
            mag.get_sensor_description(),
            data_df,
            SensorType.MAGNETOMETER,
            sample_rate,
            sample_interval,
            sample_interval_std,
            False,
        )


def load_apim_magnetometer_from_list(wrapped_packets: List[WrappedRedvoxPacketM]) -> Optional[SensorData]:
    """
    load magnetometer data from a list of wrapped packets
    :param wrapped_packets: packets with data to load
    :return: magnetometer sensor data if it exists, None otherwise
    """
    data_df = [[], [], [], []]
    for packet in wrapped_packets:
        mag = packet.get_sensors().get_magnetometer()
        if mag and packet.get_sensors().validate_magnetometer():
            data_df[0].extend(mag.get_timestamps().get_timestamps())
            data_df[1].extend(mag.get_x_samples().get_values())
            data_df[2].extend(mag.get_y_samples().get_values())
            data_df[3].extend(mag.get_z_samples().get_values())
    if len(data_df[0]) > 0:
        return load_apim_xyz_sensor(SensorType.MAGNETOMETER, data_df, "magnetometer",
                                    wrapped_packets[0].get_sensors().get_magnetometer().get_sensor_description())
    return None


def load_apim_gyroscope(wrapped_packet: WrappedRedvoxPacketM) -> Optional[SensorData]:
    """
    load gyroscope data from a single wrapped packet
    :param wrapped_packet: packet with data to load
    :return: gyroscope sensor data if it exists, None otherwise
    """
    gyro = wrapped_packet.get_sensors().get_gyroscope()
    if gyro and wrapped_packet.get_sensors().validate_gyroscope():
        data_df = read_apim_xyz_sensor(gyro, "gyroscope")
        sample_rate, sample_interval, sample_interval_std = get_sample_statistics(data_df)
        return SensorData(
            gyro.get_sensor_description(),
            data_df,
            SensorType.GYROSCOPE,
            sample_rate,
            sample_interval,
            sample_interval_std,
            False,
        )


def load_apim_gyroscope_from_list(wrapped_packets: List[WrappedRedvoxPacketM]) -> Optional[SensorData]:
    """
    load gyroscope data from a list of wrapped packets
    :param wrapped_packets: packets with data to load
    :return: gyroscope sensor data if it exists, None otherwise
    """
    data_df = [[], [], [], []]
    for packet in wrapped_packets:
        gyro = packet.get_sensors().get_gyroscope()
        if gyro and packet.get_sensors().validate_gyroscope():
            data_df[0].extend(gyro.get_timestamps().get_timestamps())
            data_df[1].extend(gyro.get_x_samples().get_values())
            data_df[2].extend(gyro.get_y_samples().get_values())
            data_df[3].extend(gyro.get_z_samples().get_values())
    if len(data_df[0]) > 0:
        return load_apim_xyz_sensor(SensorType.GYROSCOPE, data_df, "gyroscope",
                                    wrapped_packets[0].get_sensors().get_gyroscope().get_sensor_description())
    return None


def load_apim_gravity(wrapped_packet: WrappedRedvoxPacketM) -> Optional[SensorData]:
    """
    load gravity data from a single wrapped packet
    :param wrapped_packet: packet with data to load
    :return: gravity sensor data if it exists, None otherwise
    """
    gravity = wrapped_packet.get_sensors().get_gravity()
    if gravity and wrapped_packet.get_sensors().validate_gravity():
        data_df = read_apim_xyz_sensor(gravity, "gravity")
        sample_rate, sample_interval, sample_interval_std = get_sample_statistics(data_df)
        return SensorData(
            gravity.get_sensor_description(),
            data_df,
            SensorType.GRAVITY,
            sample_rate,
            sample_interval,
            sample_interval_std,
            False,
        )


def load_apim_gravity_from_list(wrapped_packets: List[WrappedRedvoxPacketM]) -> Optional[SensorData]:
    """
    load gravity data from a list of wrapped packets
    :param wrapped_packets: packets with data to load
    :return: gravity sensor data if it exists, None otherwise
    """
    data_df = [[], [], [], []]
    for packet in wrapped_packets:
        gravity = packet.get_sensors().get_gravity()
        if gravity and packet.get_sensors().validate_gravity():
            data_df[0].extend(gravity.get_timestamps().get_timestamps())
            data_df[1].extend(gravity.get_x_samples().get_values())
            data_df[2].extend(gravity.get_y_samples().get_values())
            data_df[3].extend(gravity.get_z_samples().get_values())
    if len(data_df[0]) > 0:
        return load_apim_xyz_sensor(SensorType.GRAVITY, data_df, "gravity",
                                    wrapped_packets[0].get_sensors().get_gravity().get_sensor_description())
    return None


def load_apim_orientation(wrapped_packet: WrappedRedvoxPacketM) -> Optional[SensorData]:
    """
    load orientation data from a single wrapped packet
    :param wrapped_packet: packet with data to load
    :return: orientation sensor data if it exists, None otherwise
    """
    orientation = wrapped_packet.get_sensors().get_orientation()
    if orientation and wrapped_packet.get_sensors().validate_orientation():
        data_df = read_apim_xyz_sensor(orientation, "orientation")
        sample_rate, sample_interval, sample_interval_std = get_sample_statistics(data_df)
        return SensorData(
            orientation.get_sensor_description(),
            data_df,
            SensorType.ORIENTATION,
            sample_rate,
            sample_interval,
            sample_interval_std,
            False,
        )


def load_apim_orientation_from_list(wrapped_packets: List[WrappedRedvoxPacketM]) -> Optional[SensorData]:
    """
    load orientation data from a list of wrapped packets
    :param wrapped_packets: packets with data to load
    :return: orientation sensor data if it exists, None otherwise
    """
    data_df = [[], [], [], []]
    for packet in wrapped_packets:
        orient = packet.get_sensors().get_orientation()
        if orient and packet.get_sensors().validate_orientation():
            data_df[0].extend(orient.get_timestamps().get_timestamps())
            data_df[1].extend(orient.get_x_samples().get_values())
            data_df[2].extend(orient.get_y_samples().get_values())
            data_df[3].extend(orient.get_z_samples().get_values())
    if len(data_df[0]) > 0:
        return load_apim_xyz_sensor(SensorType.ORIENTATION, data_df, "orientation",
                                    wrapped_packets[0].get_sensors().get_orientation().get_sensor_description())
    return None


def load_apim_linear_accel(wrapped_packet: WrappedRedvoxPacketM) -> Optional[SensorData]:
    """
    load linear acceleration data from a single wrapped packet
    :param wrapped_packet: packet with data to load
    :return: linear acceleration sensor data if it exists, None otherwise
    """
    linear_accel = wrapped_packet.get_sensors().get_linear_acceleration()
    if linear_accel and wrapped_packet.get_sensors().validate_linear_acceleration():
        data_df = read_apim_xyz_sensor(linear_accel, "linear_accel")
        sample_rate, sample_interval, sample_interval_std = get_sample_statistics(data_df)
        return SensorData(
            linear_accel.get_sensor_description(),
            data_df,
            SensorType.LINEAR_ACCELERATION,
            sample_rate,
            sample_interval,
            sample_interval_std,
            False,
        )


def load_apim_linear_accel_from_list(wrapped_packets: List[WrappedRedvoxPacketM]) -> Optional[SensorData]:
    """
    load linear acceleration data from a list of wrapped packets
    :param wrapped_packets: packets with data to load
    :return: linear acceleration sensor data if it exists, None otherwise
    """
    data_df = [[], [], [], []]
    for packet in wrapped_packets:
        lin_acc = packet.get_sensors().get_linear_acceleration()
        if lin_acc and packet.get_sensors().validate_linear_acceleration():
            data_df[0].extend(lin_acc.get_timestamps().get_timestamps())
            data_df[1].extend(lin_acc.get_x_samples().get_values())
            data_df[2].extend(lin_acc.get_y_samples().get_values())
            data_df[3].extend(lin_acc.get_z_samples().get_values())
    if len(data_df[0]) > 0:
        return load_apim_xyz_sensor(SensorType.LINEAR_ACCELERATION, data_df, "linear_accel",
                                    wrapped_packets[0].get_sensors().get_linear_acceleration().get_sensor_description())
    return None


def load_apim_rotation_vector(wrapped_packet: WrappedRedvoxPacketM) -> Optional[SensorData]:
    """
    load rotation vector data from a single wrapped packet
    :param wrapped_packet: packet with data to load
    :return: rotation vector sensor data if it exists, None otherwise
    """
    rotation = wrapped_packet.get_sensors().get_rotation_vector()
    if rotation and wrapped_packet.get_sensors().validate_rotation_vector():
        data_df = read_apim_xyz_sensor(rotation, "rotation_vector")
        sample_rate, sample_interval, sample_interval_std = get_sample_statistics(data_df)
        return SensorData(
            rotation.get_sensor_description(),
            data_df,
            SensorType.ROTATION_VECTOR,
            sample_rate,
            sample_interval,
            sample_interval_std,
            False,
        )


def load_apim_rotation_vector_from_list(wrapped_packets: List[WrappedRedvoxPacketM]) -> Optional[SensorData]:
    """
    load rotation vector data from a list of wrapped packets
    :param wrapped_packets: packets with data to load
    :return: rotation vector sensor data if it exists, None otherwise
    """
    data_df = [[], [], [], []]
    for packet in wrapped_packets:
        rot_vec = packet.get_sensors().get_rotation_vector()
        if rot_vec and packet.get_sensors().validate_rotation_vector():
            data_df[0].extend(rot_vec.get_timestamps().get_timestamps())
            data_df[1].extend(rot_vec.get_x_samples().get_values())
            data_df[2].extend(rot_vec.get_y_samples().get_values())
            data_df[3].extend(rot_vec.get_z_samples().get_values())
    if len(data_df[0]) > 0:
        return load_apim_xyz_sensor(SensorType.ROTATION_VECTOR, data_df, "rotation_vector",
                                    wrapped_packets[0].get_sensors().get_rotation_vector().get_sensor_description())
    return None


def load_apim_health(wrapped_packet: WrappedRedvoxPacketM) -> Optional[SensorData]:
    """
    load station health data from a single wrapped packet
    :param wrapped_packet: packet with data to load
    :return: station health data if it exists, None otherwise
    """
    metrics = wrapped_packet.get_station_information().get_station_metrics()
    timestamps = metrics.get_timestamps().get_timestamps()
    if len(timestamps) > 0:
        bat_samples = metrics.get_battery().get_values()
        bat_cur_samples = metrics.get_battery_current().get_values()
        temp_samples = metrics.get_temperature().get_values()
        net_samples = metrics.get_network_type().get_values()
        net_str_samples = metrics.get_network_strength().get_values()
        pow_samples = metrics.get_power_state().get_values()
        avail_ram_samples = metrics.get_available_ram().get_values()
        avail_disk_samples = metrics.get_available_disk().get_values()
        cell_samples = metrics.get_cell_service_state().get_values()
        data_for_df = []
        for i in range(len(timestamps)):
            new_entry = [
                timestamps[i],
                np.nan if len(bat_samples) < i + 1 else bat_samples[i],
                np.nan if len(bat_cur_samples) < i + 1 else bat_cur_samples[i],
                np.nan if len(temp_samples) < i + 1 else temp_samples[i],
                np.nan if len(net_samples) < i + 1 else net_samples[i],
                np.nan if len(net_str_samples) < i + 1 else net_str_samples[i],
                np.nan if len(pow_samples) < i + 1 else pow_samples[i],
                np.nan if len(avail_ram_samples) < i + 1 else avail_ram_samples[i],
                np.nan if len(avail_disk_samples) < i + 1 else avail_disk_samples[i],
                np.nan if len(cell_samples) < i + 1 else cell_samples[i],
            ]
            data_for_df.append(new_entry)
        data_df = pd.DataFrame(
            data_for_df,
            columns=[
                "timestamps",
                "battery_charge_remaining",
                "battery_current_strength",
                "internal_temp_c",
                "network_type",
                "network_strength",
                "power_state",
                "avail_ram",
                "avail_disk",
                "cell_service",
            ],
        )
        sample_rate = len(data_for_df) / wrapped_packet.get_packet_duration().total_seconds()
        return SensorData(
            "station health",
            data_df,
            SensorType.STATION_HEALTH,
            sample_rate,
            1 / sample_rate,
            np.nan if len(data_for_df) <= 1
            else dtu.microseconds_to_seconds(float(np.std(np.diff(data_df["timestamps"])))),
            False,
        )
    return None


def load_apim_health_from_list(wrapped_packets: List[WrappedRedvoxPacketM]) -> Optional[SensorData]:
    """
    load station health data from a list of wrapped packets
    :param wrapped_packets: packets with data to load
    :return: station health sensor data if it exists, None otherwise
    """
    data_for_df = [[], [], [], [], [], [], [], [], [], []]
    for packet in wrapped_packets:
        metrics = packet.get_station_information().get_station_metrics()
        timestamps = metrics.get_timestamps().get_timestamps()
        if len(timestamps) > 0:
            data_for_df[0].extend(timestamps)
            bat_samples = metrics.get_battery().get_values()
            if len(bat_samples) <= 0:
                bat_samples = np.full(len(timestamps), np.nan)
            data_for_df[1].extend(bat_samples)
            bat_cur_samples = metrics.get_battery_current().get_values()
            if len(bat_cur_samples) <= 0:
                bat_cur_samples = np.full(len(timestamps), np.nan)
            data_for_df[2].extend(bat_cur_samples)
            temp_samples = metrics.get_temperature().get_values()
            if len(temp_samples) <= 0:
                temp_samples = np.full(len(timestamps), np.nan)
            data_for_df[3].extend(temp_samples)
            net_samples = metrics.get_network_type().get_values()
            if len(net_samples) <= 0:
                net_samples = np.full(len(timestamps), np.nan)
            data_for_df[4].extend(net_samples)
            net_str_samples = metrics.get_network_strength().get_values()
            if len(net_str_samples) <= 0:
                net_str_samples = np.full(len(timestamps), np.nan)
            data_for_df[5].extend(net_str_samples)
            pow_samples = metrics.get_power_state().get_values()
            if len(pow_samples) <= 0:
                pow_samples = np.full(len(timestamps), np.nan)
            data_for_df[6].extend(pow_samples)
            avail_ram_samples = metrics.get_available_ram().get_values()
            if len(avail_ram_samples) <= 0:
                avail_ram_samples = np.full(len(timestamps), np.nan)
            data_for_df[7].extend(avail_ram_samples)
            avail_disk_samples = metrics.get_available_disk().get_values()
            if len(avail_disk_samples) <= 0:
                avail_disk_samples = np.full(len(timestamps), np.nan)
            data_for_df[8].extend(avail_disk_samples)
            cell_samples = metrics.get_cell_service_state().get_values()
            if len(cell_samples) <= 0:
                cell_samples = np.full(len(timestamps), np.nan)
            data_for_df[9].extend(cell_samples)
    if len(data_for_df[0]) > 0:
        # sample_rate = len(data_for_df[0]) / wrapped_packets[0].get_packet_duration().total_seconds()
        return SensorData(
            "station health",
            pd.DataFrame(
                np.transpose(data_for_df),
                columns=[
                    "timestamps",
                    "battery_charge_remaining",
                    "battery_current_strength",
                    "internal_temp_c",
                    "network_type",
                    "network_strength",
                    "power_state",
                    "avail_ram",
                    "avail_disk",
                    "cell_service",
                ],
            ),
            SensorType.STATION_HEALTH,
            calculate_stats=True
            # sample_rate,
            # 1 / sample_rate,
            # np.nan if len(data_for_df[0]) <= 1
            # else dtu.microseconds_to_seconds(float(np.std(np.diff(data_for_df[0])))),
            # False,
            )
    return None



# todo scrap if not needed
# """
# This module loads or reads station data from various sources
# """
# from dataclasses import dataclass
# import os
# import glob
# from typing import List, Dict, Optional
#
# import numpy as np
# import pandas as pd
# from obspy import read
#
# from redvox.api900 import reader as api900_io
# from redvox.common import file_statistics as fs, date_time_utils as dtu, timesync as ts
# from redvox.common.sensor_data import SensorType, SensorData
# from redvox.common.station import Station
# from redvox.common.station_utils import DataPacket, StationTiming, StationLocation, StationMetadata, LocationData, \
#     TimeSyncData
# from redvox.api1000.wrapped_redvox_packet.sensors.location import LocationProvider
# from redvox.api1000.wrapped_redvox_packet.station_information import NetworkType
# from redvox.api1000.wrapped_redvox_packet.sensors import xyz, single
# from redvox.api1000.wrapped_redvox_packet import wrapped_packet as apim_wp
#
#
# @dataclass
# class StationSummary:
#     """
#     Contains a summary of each stations' data reader results.
#     properties:
#         station_id: str, station id
#         station_uuid: str, station uuid
#         os: str, station os
#         os_version: str, station os version
#         app_version: str, station app version
#         audio_sampling_rate_hz: float, sample rate in hz
#         total_duration_s: float, duration of data in seconds
#         start_dt: dtu.datetime object, start datetime of data read
#         end_dt: dtu.datetime object, end datetime of data read
#     """
#     station_id: str
#     station_uuid: str
#     # pylint: ignore=C0103
#     os: str
#     os_version: str
#     app_version: str
#     audio_sampling_rate_hz: float
#     total_duration_s: float
#     start_dt: dtu.datetime
#     end_dt: dtu.datetime
#
#     @staticmethod
#     def from_station(station: Station) -> 'StationSummary':
#         """
#         :param station: the station to make a summary for
#         :return: the station summary of a single station
#         """
#         total_duration: float = station.audio_sensor().data_duration_s()
#         start_dt: dtu.datetime = dtu.datetime_from_epoch_microseconds_utc(station.audio_sensor().first_data_timestamp())
#         end_dt: dtu.datetime = dtu.datetime_from_epoch_microseconds_utc(station.audio_sensor().last_data_timestamp())
#
#         station_info = station.station_metadata
#         audio = station.audio_sensor()
#         return StationSummary(
#             station_info.station_id,
#             station_info.station_uuid,
#             station_info.station_os if station_info.station_os is not None else "OS UNKNOWN",
#             station_info.station_os_version if station_info.station_os_version is not None else "OS VERSION UNKNOWN",
#             station_info.station_app_version if station_info.station_app_version is not None else "APP VERSION UNKNOWN",
#             audio.sample_rate if audio is not None else np.nan,
#             total_duration,
#             start_dt,
#             end_dt
#         )
#
#
# class ReadResult:
#     """
#     Stores station information after being read from files
#     Properties:
#         station_id_uuid_to_stations: dict of string to Station object, where the string is id:uuid format
#         __station_id_to_id_uuid: dict of string to string, maps id to uuid
#         __station_summaries: dict of string to StationSummary object, maps id to StationSummary
#     """
#     def __init__(self,
#                  station_id_uuid_to_stations: Dict[str, Station]):
#         # result_stations: List[Station] = None):
#         """
#         :param station_id_uuid_to_stations: station_id:station_uuid -> station information
#         """
#         # station_keys: List[StationKey] = []
#         # self.station_keys_to_stations: Dict[StationKey, Station]
#         # for station in result_stations:
#         #     station_keys.append(station.station_key)
#         #     self.station_keys_to_stations[station.station_key] = station
#         # todo: add mach time zero as another key
#         self.station_id_uuid_to_stations: Dict[str, Station] = station_id_uuid_to_stations
#         self.__station_id_to_id_uuid: Dict[str, str] = {}
#         self.__station_summaries: Dict[str, StationSummary] = {}
#
#         self._update_metadata()
#
#     # def search_stations(self, id_or_uuid: str, start_time_micros: float = np.inf) -> Station:
#     #     """
#     #     Searches the stations in the ReadResult for the first one that matches the exact id or uuid given,
#     #         with starting time on or before the time given
#     #     :param id_or_uuid: str, the id or uuid of the station to search for
#     #     :param start_time_micros: float, the start time of the station to search for, default np.inf
#     #     :return: Station object that has the id or uuid and was started on or before the time given
#     #     """
#     #
#     # def get_stations_by_id(self, id_or_uuid: str) -> List[Station]:
#     #     pass
#     #     return False
#     #
#     # def get_stations_by_start_time(self, start_time_micros: float) -> List[Station]:
#     #     pass
#     #     return False
#
#     def _update_metadata(self):
#         """
#         updates ids:uuids pairs and summary information
#         """
#         for id_uuid, station in self.station_id_uuid_to_stations.items():
#             s: List[str] = id_uuid.split(":")
#             self.__station_id_to_id_uuid[s[0]] = s[1]
#             self.__station_summaries[s[0]] = StationSummary.from_station(station)
#
#     def __get_station_id_by_uuid(self, uuid: str) -> str:
#         """
#         given a uuid, returns the station_id
#         :param uuid: uuid to search for
#         :return: the station_id of the uuid, or an empty string if uuid doesn't exist
#         """
#         for station_id, station_uuid in self.__station_id_to_id_uuid.items():
#             if station_uuid == uuid:
#                 return station_id
#         return ""
#
#     def __get_station_id(self, string_id: str) -> str:
#         """
#         given an id, uuid, or id:uuid, return id
#         :param string_id: string, the id to examine
#         :return: the station id corresponding to the input or an empty string if it doesn't exist
#         """
#         if self.check_for_id(string_id):
#             if string_id in self.station_id_uuid_to_stations.keys():
#                 s: List[str] = string_id.split(":")
#                 return s[0]
#             elif string_id in self.__station_id_to_id_uuid.values():
#                 return self.__get_station_id_by_uuid(string_id)
#             elif string_id in self.__station_id_to_id_uuid.keys():
#                 return string_id
#         return ""
#
#     def pop_station(self, station_id: str) -> 'ReadResult':
#         """
#         removes a station from the ReadResult; station_id can be one of id, uuid or id:uuid
#         :param station_id: station to remove
#         :return: copy of ReadResult without the station_id specified
#         """
#         if ":" in station_id:
#             s: List[str] = station_id.split(":")
#             station_id = s[0]
#         elif station_id in self.__station_id_to_id_uuid.values():
#             station_id = self.__get_station_id_by_uuid(station_id)
#         if self.check_for_id(station_id):  # the station_id has been converted into id from id:uuid or uuid
#             self.station_id_uuid_to_stations.pop(f"{station_id}:{self.__station_id_to_id_uuid[station_id]}")
#             self.__station_id_to_id_uuid.pop(station_id)
#             self.__station_summaries.pop(station_id)
#         else:
#             print(f"ReadResult cannot remove station {station_id} because it does not exist")
#         return self
#
#     def check_for_id(self, check_id: str) -> bool:
#         """
#         Look at keys and shortened keys in for the check_id; must be one of id, id:uuid, or uuid
#         :param check_id: id to look for
#         :return: True if check_id is in the ReadResult
#         """
#         return check_id in self.__station_id_to_id_uuid.keys() or \
#             check_id in self.__station_id_to_id_uuid.values() or check_id in self.station_id_uuid_to_stations.keys()
#
#     def get_station(self, station_id: str) -> Optional[Station]:
#         """
#         Find the station identified by the station_id given; it can be id or id:uuid
#         :param station_id: str id of station; can be id or id:uuid
#         :return: the station if it exists, None otherwise
#         """
#         if ":" in station_id:
#             return self.station_id_uuid_to_stations[station_id]
#         elif self.check_for_id(station_id):
#             return self.station_id_uuid_to_stations[f"{station_id}:{self.__station_id_to_id_uuid[station_id]}"]
#         print(f"WARNING: ReadResult attempted to read station id: {station_id}, but could not find id in results")
#         return None
#
#     def get_all_stations(self) -> List[Station]:
#         """
#         :return: a list of all stations in the ReadResult
#         """
#         return list(self.station_id_uuid_to_stations.values())
#
#     def get_all_station_ids(self) -> List[str]:
#         """
#         :return: a list of all station ids in the ReadResult
#         """
#         return list(self.__station_id_to_id_uuid.keys())
#
#     def get_station_summary(self, station_id: str) -> Optional[StationSummary]:
#         """
#         Find the station summary identified by the station_id given; it can be id or id:uuid
#         :return: A StationSummary in this ReadResult if it exists, None otherwise
#         """
#         if ":" in station_id:
#             s: List[str] = station_id.split(":")
#             return self.__station_summaries[s[0]]
#         elif self.check_for_id(station_id):
#             return self.__station_summaries[self.__get_station_id(station_id)]
#         print(f"WARNING: ReadResult attempted to read station id: {station_id}, but could not find id in results")
#         return None
#
#     def get_station_summaries(self) -> List[StationSummary]:
#         """
#         :return: A list of StationSummaries contained in this ReadResult
#         """
#         return list(self.__station_summaries.values())
#
#     def append_station(self, new_station_id: str, new_station: Station):
#         """
#         adds a station to the ReadResult.  Appends data to existing stations
#         :param new_station_id: id of station to add
#         :param new_station: Station object to add
#         """
#         if self.check_for_id(new_station_id):
#             self.station_id_uuid_to_stations[new_station_id].append_station(new_station)
#         else:
#             self.station_id_uuid_to_stations[new_station_id] = new_station
#             self.__station_id_to_id_uuid[new_station.station_metadata.station_id] = \
#                 new_station.station_metadata.station_uuid
#             self.__station_summaries[new_station.station_metadata.station_id] = \
#                 (StationSummary.from_station(new_station))
#
#     def append(self, new_stations: 'ReadResult'):
#         """
#         adds stations from another ReadResult to the calling ReadResult
#         :param new_stations: ReadResult object with stations to add
#         """
#         for new_station_id, new_station in new_stations.station_id_uuid_to_stations.items():
#             self.append_station(new_station_id, new_station)
#
#
# def calc_evenly_sampled_timestamps(start: float, samples: int, rate_hz: float) -> np.array:
#     """
#     given a start time, calculates samples amount of evenly spaced timestamps at rate_hz
#     :param start: float, start timestamp in microseconds
#     :param samples: int, number of samples
#     :param rate_hz: float, sample rate in hz
#     :return: np.array with evenly spaced timestamps starting at start
#     """
#     return start + dtu.seconds_to_microseconds(np.arange(0, samples) / rate_hz)
#
#
# def read_api900_non_mic_sensor(sensor: api900_io.RedvoxSensor, column_id: str) -> SensorData:
#     """
#     read a sensor that does not have mic data from an api900 data packet
#     :param sensor: the non-mic api900 sensor to read
#     :param column_id: string, used to name the columns
#     :return: generic SensorData object
#     """
#     timestamps = sensor.timestamps_microseconds_utc()
#     if len(timestamps) > 1:
#         sample_interval = dtu.microseconds_to_seconds(float(np.mean(np.diff(timestamps))))
#         sample_interval_std = dtu.microseconds_to_seconds(float(np.std(np.diff(timestamps))))
#     else:
#         sample_interval = np.nan
#         sample_interval_std = np.nan
#     if type(sensor) in [api900_io.AccelerometerSensor, api900_io.MagnetometerSensor, api900_io.GyroscopeSensor]:
#         data_for_df = np.transpose([timestamps,
#                                     sensor.payload_values_x(), sensor.payload_values_y(), sensor.payload_values_z()])
#         columns = ["timestamps", f"{column_id}_x", f"{column_id}_y", f"{column_id}_z"]
#     else:
#         data_for_df = np.transpose([timestamps, sensor.payload_values()])
#         columns = ["timestamps", column_id]
#     return SensorData(sensor.sensor_name(), pd.DataFrame(data_for_df, columns=columns), 1 / sample_interval,
#                       sample_interval, sample_interval_std, False)
#
#
# def read_api900_wrapped_packet(wrapped_packet: api900_io.WrappedRedvoxPacket) -> Dict[SensorType, SensorData]:
#     """
#     reads the data from a wrapped api900 redvox packet into a dictionary of generic data
#     :param wrapped_packet: a wrapped api900 redvox packet
#     :return: a dictionary containing all the sensor data
#     """
#     data_dict: Dict[SensorType, SensorData] = {}
#     # there are 9 api900 sensors
#     if wrapped_packet.has_microphone_sensor():
#         sample_rate_hz = wrapped_packet.microphone_sensor().sample_rate_hz()
#         timestamps = calc_evenly_sampled_timestamps(
#             wrapped_packet.microphone_sensor().first_sample_timestamp_epoch_microseconds_utc(),
#             fs.get_num_points_from_sample_rate(sample_rate_hz), sample_rate_hz)
#         data_for_df = np.transpose([timestamps, wrapped_packet.microphone_sensor().payload_values().astype(float)])
#         data_dict[SensorType.AUDIO] = SensorData(wrapped_packet.microphone_sensor().sensor_name(),
#                                                  pd.DataFrame(data_for_df, columns=["timestamps", "microphone"]),
#                                                  sample_rate_hz, 1 / sample_rate_hz, 0.0, True)
#     if wrapped_packet.has_accelerometer_sensor():
#         data_dict[SensorType.ACCELEROMETER] = \
#             read_api900_non_mic_sensor(wrapped_packet.accelerometer_sensor(), "accelerometer")
#     if wrapped_packet.has_magnetometer_sensor():
#         data_dict[SensorType.MAGNETOMETER] = \
#             read_api900_non_mic_sensor(wrapped_packet.magnetometer_sensor(), "magnetometer")
#     if wrapped_packet.has_gyroscope_sensor():
#         data_dict[SensorType.GYROSCOPE] = read_api900_non_mic_sensor(wrapped_packet.gyroscope_sensor(), "gyroscope")
#     if wrapped_packet.has_barometer_sensor():
#         data_dict[SensorType.PRESSURE] = read_api900_non_mic_sensor(wrapped_packet.barometer_sensor(), "barometer")
#     if wrapped_packet.has_light_sensor():
#         data_dict[SensorType.LIGHT] = read_api900_non_mic_sensor(wrapped_packet.light_sensor(), "light")
#     if wrapped_packet.has_infrared_sensor():
#         data_dict[SensorType.INFRARED] = read_api900_non_mic_sensor(wrapped_packet.infrared_sensor(), "infrared")
#     if wrapped_packet.has_image_sensor():
#         data_dict[SensorType.IMAGE] = read_api900_non_mic_sensor(wrapped_packet.image_sensor(), "image")
#     if wrapped_packet.has_location_sensor():
#         timestamps = wrapped_packet.location_sensor().timestamps_microseconds_utc()
#         if len(timestamps) > 1:
#             sample_interval = dtu.microseconds_to_seconds(float(np.mean(np.diff(timestamps))))
#             sample_interval_std = dtu.microseconds_to_seconds(float(np.std(np.diff(timestamps))))
#         else:
#             sample_interval = np.nan
#             sample_interval_std = np.nan
#         if wrapped_packet.location_sensor().check_for_preset_lat_lon():
#             lat_lon = wrapped_packet.location_sensor().get_payload_lat_lon()
#             data_for_df = np.array([[timestamps[0], lat_lon[0], lat_lon[1], np.nan, np.nan, np.nan,
#                                      LocationProvider.USER, np.nan, np.nan, np.nan, np.nan]])
#         else:
#             if wrapped_packet.location_sensor().sensor_name().lower() == "network":
#                 provider = LocationProvider.NETWORK
#             elif wrapped_packet.location_sensor().sensor_name().lower() == "gps":
#                 provider = LocationProvider.GPS
#             else:
#                 provider = LocationProvider.UNKNOWN
#             data_for_df = np.transpose([timestamps,
#                                         wrapped_packet.location_sensor().payload_values_latitude(),
#                                         wrapped_packet.location_sensor().payload_values_longitude(),
#                                         wrapped_packet.location_sensor().payload_values_altitude(),
#                                         wrapped_packet.location_sensor().payload_values_speed(),
#                                         wrapped_packet.location_sensor().payload_values_accuracy(),
#                                         np.full(len(timestamps), provider),
#                                         np.full(len(timestamps), np.nan), np.full(len(timestamps), np.nan),
#                                         np.full(len(timestamps), np.nan), np.full(len(timestamps), np.nan)])
#         columns = ["timestamps", "latitude", "longitude", "altitude", "speed", "horizontal_accuracy",
#                    "location_provider", "bearing", "vertical_accuracy", "speed_accuracy", "bearing_accuracy"]
#         data_dict[SensorType.LOCATION] = SensorData(wrapped_packet.location_sensor().sensor_name(),
#                                                     pd.DataFrame(data_for_df, columns=columns),
#                                                     1 / sample_interval, sample_interval, sample_interval_std, False)
#     packet_duration_s = wrapped_packet.duration_s()
#     if wrapped_packet.has_time_synchronization_sensor():
#         network_type = NetworkType.UNKNOWN_NETWORK
#     else:
#         network_type = NetworkType.NO_NETWORK
#     data_dict[SensorType.STATION_HEALTH] = SensorData("station health",
#                                                       pd.DataFrame([[wrapped_packet.start_timestamp_us_utc(),
#                                                                      wrapped_packet.battery_level_percent(),
#                                                                      wrapped_packet.device_temperature_c(),
#                                                                      network_type]],
#                                                                    columns=["timestamps", "battery_charge_remaining",
#                                                                             "internal_temp_c", "network_type"]),
#                                                       1/packet_duration_s, packet_duration_s, 0, False)
#     return data_dict
#
#
# def load_station_from_api900(api900_packet: api900_io.WrappedRedvoxPacket,
#                              start_timestamp_utc_s: Optional[int] = None,
#                              end_timestamp_utc_s: Optional[int] = None) -> Station:
#     """
#     reads in station data from a single wrapped api900 packet
#     :param api900_packet: wrapped api900 packet to read from
#     :param start_timestamp_utc_s: The start timestamp as seconds since the epoch UTC.
#     :param end_timestamp_utc_s: The end timestamp as seconds since the epoch UTC.
#     :return: a station Object
#     """
#     # set station metadata and timing
#     timing = StationTiming(api900_packet.mach_time_zero(), api900_packet.microphone_sensor().sample_rate_hz(),
#                            api900_packet.microphone_sensor().first_sample_timestamp_epoch_microseconds_utc(),
#                            start_timestamp_utc_s, end_timestamp_utc_s,
#                            np.nan if api900_packet.best_latency() is None else api900_packet.best_latency(),
#                            0.0 if api900_packet.best_offset() is None else api900_packet.best_offset())
#     metadata = StationMetadata(api900_packet.redvox_id(), api900_packet.device_make(),
#                                api900_packet.device_model(), False, api900_packet.device_os(),
#                                api900_packet.device_os_version(), "Redvox", api900_packet.app_version(),
#                                api900_packet.is_scrambled(), timing, station_uuid=api900_packet.uuid())
#     data_dict = read_api900_wrapped_packet(api900_packet)
#     timesync_data = TimeSyncData(api900_packet.time_synchronization_sensor().payload_values(),
#                                  np.nan if api900_packet.best_latency() is None else api900_packet.best_latency(),
#                                  0.0 if api900_packet.best_offset() is None else api900_packet.best_offset())
#     packet_data = DataPacket(api900_packet.server_timestamp_epoch_microseconds_utc(),
#                              api900_packet.app_file_start_timestamp_machine(),
#                              float(api900_packet.start_timestamp_us_utc()), api900_packet.end_timestamp_us_utc(),
#                              len(api900_packet.microphone_sensor().payload_values()) /
#                              api900_packet.microphone_sensor().sample_rate_hz(),
#                              len(api900_packet.microphone_sensor().payload_values()), timesync_data)
#     packet_list: List[DataPacket] = [packet_data]
#     # get the best timing values for the station
#     if timing.station_best_latency is None or np.isnan(timing.station_best_latency):
#         ts_analysis = ts.TimeSyncData(packet_data, metadata)
#         metadata.timing_data.station_best_latency = ts_analysis.best_latency
#         metadata.timing_data.station_best_offset = ts_analysis.best_offset
#         metadata.timing_data.station_mean_offset = ts_analysis.mean_offset
#         metadata.timing_data.station_std_offset = ts_analysis.offset_std
#     return Station(metadata, data_dict, packet_list)
#
#
# def load_station_from_api900_file(directory: str, start_timestamp_utc_s: Optional[int] = None,
#                                   end_timestamp_utc_s: Optional[int] = None) -> Station:
#     """
#     reads in station data from a single api900 file
#     :param directory: string of the file to read from
#     :param start_timestamp_utc_s: The start timestamp as seconds since the epoch UTC.
#     :param end_timestamp_utc_s: The end timestamp as seconds since the epoch UTC.
#     :return: a station Object
#     """
#     api900_packet = api900_io.read_rdvxz_file(directory)
#     return load_station_from_api900(api900_packet, start_timestamp_utc_s, end_timestamp_utc_s)
#
#
# def load_file_range_from_api900(directory: str,
#                                 start_timestamp_utc_s: Optional[int] = None,
#                                 end_timestamp_utc_s: Optional[int] = None,
#                                 redvox_ids: Optional[List[str]] = None,
#                                 structured_layout: bool = False,
#                                 concat_continuous_segments: bool = True) -> ReadResult:
#     """
#     reads in api900 data from a directory and returns a list of stations
#     note that the param descriptions are taken directly from api900.reader.read_rdvxz_file_range
#     :param directory: The root directory of the data. If structured_layout is False, then this directory will
#                       contain various unorganized .rdvxz files. If structured_layout is True, then this directory
#                       must be the root api900 directory of the structured files.
#     :param start_timestamp_utc_s: The start timestamp as seconds since the epoch UTC.
#     :param end_timestamp_utc_s: The end timestamp as seconds since the epoch UTC.
#     :param redvox_ids: An optional list of redvox_ids to filter against (default=[]).
#     :param structured_layout: An optional value to define if this is loading structured data (default=False).
#     :param concat_continuous_segments: An optional value to define if this function should concatenate rdvxz files
#                                        into multiple continuous rdvxz files separated at gaps.
#     :return: a list of Station objects that contain the data
#     """
#     all_stations: ReadResult = ReadResult({})
#     all_data = api900_io.read_rdvxz_file_range(directory, start_timestamp_utc_s, end_timestamp_utc_s, redvox_ids,
#                                                structured_layout, concat_continuous_segments)
#     for redvox_id, wrapped_packets in all_data.items():
#         # set station metadata and timing based on first packet
#         timing = StationTiming(wrapped_packets[0].mach_time_zero(),
#                                wrapped_packets[0].microphone_sensor().sample_rate_hz(),
#                                wrapped_packets[0].microphone_sensor().first_sample_timestamp_epoch_microseconds_utc(),
#                                start_timestamp_utc_s, end_timestamp_utc_s,
#                                np.nan if wrapped_packets[0].best_latency() is None else
#                                wrapped_packets[0].best_latency(),
#                                0.0 if wrapped_packets[0].best_offset() is None else wrapped_packets[0].best_offset())
#         metadata = StationMetadata(wrapped_packets[0].redvox_id(), wrapped_packets[0].device_make(),
#                                    wrapped_packets[0].device_model(), False, wrapped_packets[0].device_os(),
#                                    wrapped_packets[0].device_os_version(), "Redvox", wrapped_packets[0].app_version(),
#                                    wrapped_packets[0].is_scrambled(), timing, station_uuid=wrapped_packets[0].uuid())
#         # add data from packets
#         new_station = Station(metadata)
#         packet_list: List[DataPacket] = []
#         for packet in wrapped_packets:
#             if packet.has_time_synchronization_sensor():
#                 time_sync = packet.time_synchronization_sensor().payload_values()
#             else:
#                 time_sync = np.array([])
#             timesync_data = TimeSyncData(time_sync,
#                                          np.nan if packet.best_latency() is None else packet.best_latency(),
#                                          0.0 if packet.best_offset() is None else packet.best_offset())
#             data_dict = read_api900_wrapped_packet(packet)
#             new_station.append_station_data(data_dict)
#             packet_data = DataPacket(packet.server_timestamp_epoch_microseconds_utc(),
#                                      packet.app_file_start_timestamp_machine(),
#                                      float(packet.start_timestamp_us_utc()), packet.end_timestamp_us_utc(),
#                                      data_dict[SensorType.AUDIO].data_duration_s(),
#                                      data_dict[SensorType.AUDIO].num_samples(),
#                                      timesync_data)
#             packet_list.append(packet_data)
#         new_station.packet_data = packet_list
#
#         if timing.station_best_latency is None or np.isnan(timing.station_best_latency):
#             # get the best timing values for the station
#             ts_analysis = ts.TimeSyncAnalysis(new_station)
#             new_station.station_metadata.timing_data.station_best_latency = ts_analysis.get_best_latency()
#             new_station.station_metadata.timing_data.station_best_offset = ts_analysis.get_best_offset()
#             new_station.station_metadata.timing_data.station_mean_offset = ts_analysis.get_mean_offset()
#             new_station.station_metadata.timing_data.station_std_offset = ts_analysis.get_offset_stdev()
#
#         # create the Station data object
#         all_stations.append_station(redvox_id, new_station)
#
#     return all_stations
#
#
# def read_apim_xyz_sensor(sensor: xyz.Xyz, column_id: str) -> SensorData:
#     """
#     read a sensor that has xyz data channels from an api M data packet
#     raises Attribute Error if sensor does not contain xyz channels
#     :param sensor: the xyz api M sensor to read
#     :param column_id: string, used to name the columns
#     :return: generic SensorData representation of the xyz channel sensor
#     """
#     timestamps = sensor.get_timestamps().get_timestamps()
#     if len(timestamps) > 1:
#         sample_interval = dtu.microseconds_to_seconds(float(np.mean(np.diff(timestamps))))
#         sample_interval_std = dtu.microseconds_to_seconds(float(np.std(np.diff(timestamps))))
#     else:
#         sample_interval = np.nan
#         sample_interval_std = np.nan
#     try:
#         data_for_df = np.transpose([timestamps,
#                                     sensor.get_x_samples().get_values(),
#                                     sensor.get_y_samples().get_values(),
#                                     sensor.get_z_samples().get_values()])
#         columns = ["timestamps", f"{column_id}_x", f"{column_id}_y", f"{column_id}_z"]
#         return SensorData(sensor.get_sensor_description(), pd.DataFrame(data_for_df, columns=columns),
#                           1 / sample_interval, sample_interval, sample_interval_std, False)
#     except AttributeError:
#         raise
#
#
# def read_apim_single_sensor(sensor: single.Single, column_id: str) -> SensorData:
#     """
#     read a sensor that has a single data channel from an api M data packet
#     raises Attribute Error if sensor does not contain exactly one
#     :param sensor: the single channel api M sensor to read
#     :param column_id: string, used to name the columns
#     :return: generic SensorData representation of the single channel sensor
#     """
#     timestamps = sensor.get_timestamps().get_timestamps()
#     if len(timestamps) > 1:
#         sample_interval = dtu.microseconds_to_seconds(float(np.mean(np.diff(timestamps))))
#         sample_interval_std = dtu.microseconds_to_seconds(float(np.std(np.diff(timestamps))))
#     else:
#         sample_interval = np.nan
#         sample_interval_std = np.nan
#     try:
#         data_for_df = np.transpose([timestamps, sensor.get_samples().get_values()])
#         columns = ["timestamps", column_id]
#         return SensorData(sensor.get_sensor_description(), pd.DataFrame(data_for_df, columns=columns),
#                           1 / sample_interval, sample_interval, sample_interval_std, False)
#     except AttributeError:
#         raise
#
#
# def load_apim_wrapped_packet(wrapped_packet: apim_wp.WrappedRedvoxPacketM) -> Dict[SensorType, SensorData]:
#     """
#     reads the data from a wrapped api M redvox packet into a dictionary of generic data
#     :param wrapped_packet: a wrapped api M redvox packet
#     :return: a dictionary containing all the sensor data
#     """
#     data_dict: Dict[SensorType, SensorData] = {}
#     sensors = wrapped_packet.get_sensors()
#     # there are 17 api M sensors
#     if sensors.has_audio() and sensors.validate_audio():
#         sample_rate_hz = sensors.get_audio().get_sample_rate()
#         data_for_df = sensors.get_audio().get_samples().get_values()
#         timestamps = calc_evenly_sampled_timestamps(sensors.get_audio().get_first_sample_timestamp(),
#                                                     len(data_for_df), sample_rate_hz)
#         data_dict[SensorType.AUDIO] = SensorData(sensors.get_audio().get_sensor_description(),
#                                                  pd.DataFrame(np.transpose([timestamps, data_for_df]),
#                                                               columns=["timestamps", "microphone"]),
#                                                  sample_rate_hz, 1 / sample_rate_hz, 0.0, True)
#     if sensors.has_compressed_audio() and sensors.validate_compressed_audio():
#         sample_rate_hz = sensors.get_compressed_audio().get_sample_rate()
#         data_for_df = sensors.get_compressed_audio().get_samples().get_values()
#         timestamps = calc_evenly_sampled_timestamps(sensors.get_compressed_audio().get_first_sample_timestamp(),
#                                                     len(data_for_df), sample_rate_hz)
#         data_dict[SensorType.COMPRESSED_AUDIO] = SensorData(sensors.get_compressed_audio().get_sensor_description(),
#                                                             pd.DataFrame(np.transpose([timestamps, data_for_df]),
#                                                                          columns=["compressed_audio"]),
#                                                             sample_rate_hz, 1 / sample_rate_hz, 0.0, True)
#     if sensors.has_accelerometer() and sensors.validate_accelerometer():
#         data_dict[SensorType.ACCELEROMETER] = read_apim_xyz_sensor(sensors.get_accelerometer(), "accelerometer")
#     if sensors.has_magnetometer() and sensors.validate_magnetometer():
#         data_dict[SensorType.MAGNETOMETER] = read_apim_xyz_sensor(sensors.get_magnetometer(), "magnetometer")
#     if sensors.has_linear_acceleration() and sensors.validate_accelerometer():
#         data_dict[SensorType.LINEAR_ACCELERATION] = read_apim_xyz_sensor(sensors.get_linear_acceleration(),
#                                                                          "linear_accel")
#     if sensors.has_orientation() and sensors.validate_orientation():
#         data_dict[SensorType.ORIENTATION] = read_apim_xyz_sensor(sensors.get_orientation(), "orientation")
#     if sensors.has_rotation_vector() and sensors.validate_rotation_vector():
#         data_dict[SensorType.ROTATION_VECTOR] = read_apim_xyz_sensor(sensors.get_rotation_vector(), "rotation_vector")
#     if sensors.has_gyroscope() and sensors.validate_gyroscope():
#         data_dict[SensorType.GYROSCOPE] = read_apim_xyz_sensor(sensors.get_gyroscope(), "gyroscope")
#     if sensors.has_gravity() and sensors.validate_gravity():
#         data_dict[SensorType.GRAVITY] = read_apim_xyz_sensor(sensors.get_gravity(), "gravity")
#     if sensors.has_pressure() and sensors.validate_pressure():
#         data_dict[SensorType.PRESSURE] = read_apim_single_sensor(sensors.get_pressure(), "barometer")
#     if sensors.has_light() and sensors.validate_light():
#         data_dict[SensorType.LIGHT] = read_apim_single_sensor(sensors.get_light(), "light")
#     if sensors.has_proximity() and sensors.validate_proximity():
#         data_dict[SensorType.PROXIMITY] = read_apim_single_sensor(sensors.get_proximity(), "proximity")
#     if sensors.has_ambient_temperature() and sensors.validate_ambient_temperature():
#         data_dict[SensorType.AMBIENT_TEMPERATURE] = read_apim_single_sensor(sensors.get_ambient_temperature(),
#                                                                             "ambient_temp")
#     if sensors.has_relative_humidity() and sensors.validate_relative_humidity():
#         data_dict[SensorType.RELATIVE_HUMIDITY] = read_apim_single_sensor(sensors.get_relative_humidity(),
#                                                                           "rel_humidity")
#     if sensors.has_image() and sensors.validate_image():
#         timestamps = sensors.get_image().get_timestamps().get_timestamps()
#         if len(timestamps) > 1:
#             sample_interval = dtu.microseconds_to_seconds(float(np.mean(np.diff(timestamps))))
#             sample_interval_std = dtu.microseconds_to_seconds(float(np.std(np.diff(timestamps))))
#         else:
#             sample_interval = np.nan
#             sample_interval_std = np.nan
#         codecs = np.full(len(timestamps), sensors.get_image().get_image_codec().value)
#         data_for_df = np.transpose([timestamps, sensors.get_image().get_samples(), codecs])
#         data_dict[SensorType.IMAGE] = SensorData(sensors.get_image().get_sensor_description(),
#                                                  pd.DataFrame(data_for_df,
#                                                               columns=["timestamps", "image", "image_codec"]),
#                                                  1 / sample_interval, sample_interval, sample_interval_std, False)
#     if sensors.has_location() and sensors.validate_location():
#         if sensors.get_location().is_only_best_values():
#             sample_interval = np.nan
#             sample_interval_std = np.nan
#             if sensors.get_location().get_last_best_location():
#                 best_loc = sensors.get_location().get_last_best_location()
#             else:
#                 best_loc = sensors.get_location().get_overall_best_location()
#             data_for_df = [[best_loc.get_latitude_longitude_timestamp().get_mach(),
#                             best_loc.get_latitude(),
#                             best_loc.get_longitude(),
#                             best_loc.get_altitude(),
#                             best_loc.get_speed(),
#                             best_loc.get_bearing(),
#                             best_loc.get_horizontal_accuracy(),
#                             best_loc.get_vertical_accuracy(),
#                             best_loc.get_speed_accuracy(),
#                             best_loc.get_bearing_accuracy(),
#                             best_loc.get_location_provider()]]
#         else:
#             timestamps = sensors.get_location().get_timestamps().get_timestamps()
#             if len(timestamps) > 1:
#                 sample_interval = dtu.microseconds_to_seconds(float(np.mean(np.diff(timestamps))))
#                 sample_interval_std = dtu.microseconds_to_seconds(float(np.std(np.diff(timestamps))))
#             else:
#                 sample_interval = np.nan
#                 sample_interval_std = np.nan
#             data_for_df = np.transpose([timestamps,
#                                         sensors.get_location().get_latitude_samples().get_values(),
#                                         sensors.get_location().get_longitude_samples().get_values(),
#                                         sensors.get_location().get_altitude_samples().get_values(),
#                                         sensors.get_location().get_speed_samples().get_values(),
#                                         sensors.get_location().get_bearing_samples().get_values(),
#                                         sensors.get_location().get_horizontal_accuracy_samples().get_values(),
#                                         sensors.get_location().get_vertical_accuracy_samples().get_values(),
#                                         sensors.get_location().get_speed_samples().get_values(),
#                                         sensors.get_location().get_bearing_accuracy_samples().get_values(),
#                                         sensors.get_location().get_location_providers().get_values()])
#         columns = ["timestamps", "latitude", "longitude", "altitude", "speed", "bearing",
#                    "horizontal_accuracy", "vertical_accuracy", "speed_accuracy", "bearing_accuracy",
#                    "location_provider"]
#         data_dict[SensorType.LOCATION] = SensorData(sensors.get_location().get_sensor_description(),
#                                                     pd.DataFrame(data_for_df, columns=columns),
#                                                     1 / sample_interval, sample_interval,
#                                                     sample_interval_std, False)
#     station_metrics = wrapped_packet.get_station_information().get_station_metrics()
#     timestamps = station_metrics.get_timestamps().get_timestamps()
#     if len(timestamps) > 0:
#         if len(timestamps) > 1:
#             sample_interval = dtu.microseconds_to_seconds(float(np.mean(np.diff(timestamps))))
#             sample_interval_std = dtu.microseconds_to_seconds(float(np.std(np.diff(timestamps))))
#         else:
#             sample_interval = np.nan
#             sample_interval_std = np.nan
#         data_dict[SensorType.STATION_HEALTH] = \
#             SensorData("station health",
#                        pd.DataFrame(np.transpose([timestamps,
#                                                   station_metrics.get_battery().get_values(),
#                                                   station_metrics.get_battery_current().get_values(),
#                                                   station_metrics.get_temperature().get_values(),
#                                                   station_metrics.get_network_type().get_values(),
#                                                   station_metrics.get_network_strength().get_values(),
#                                                   station_metrics.get_power_state().get_values(),
#                                                   station_metrics.get_available_ram().get_values(),
#                                                   station_metrics.get_available_disk().get_values(),
#                                                   station_metrics.get_cell_service_state().get_values()
#                                                   ]),
#                                     columns=["timestamps", "battery_charge_remaining", "battery_current_strength",
#                                              "internal_temp_c", "network_type", "network_strength",
#                                              "power_state", "avail_ram", "avail_disk", "cell_service"]),
#                        1 / sample_interval, sample_interval_std, 0, False)
#     return data_dict
#
#
# def load_station_from_apim_file(directory: str, start_timestamp_utc_s: Optional[int] = None,
#                                 end_timestamp_utc_s: Optional[int] = None) -> Station:
#     """
#     reads in station data from a single api M file
#     :param directory: string of the file to read from
#     :param start_timestamp_utc_s: The start timestamp as seconds since the epoch UTC.
#     :param end_timestamp_utc_s: The end timestamp as seconds since the epoch UTC.
#     :return: a station Object
#     """
#     read_packet = apim_io.read_rdvxm_file(directory)
#     # set station metadata and timing based on first packet
#     if read_packet.get_sensors().validate_audio():
#         timing = StationTiming(read_packet.get_timing_information().get_app_start_mach_timestamp(),
#                                read_packet.get_sensors().get_audio().get_sample_rate(),
#                                read_packet.get_sensors().get_audio().get_first_sample_timestamp(),
#                                start_timestamp_utc_s, end_timestamp_utc_s,
#                                np.nan if read_packet.get_timing_information().get_best_latency() is None else
#                                read_packet.get_timing_information().get_best_latency(),
#                                0.0 if read_packet.get_timing_information().get_best_offset() is None else
#                                read_packet.get_timing_information().get_best_offset())
#     else:
#         raise ValueError("Station is missing Audio sensor!")
#     if read_packet.get_sensors().validate_location() and \
#             read_packet.get_sensors().get_location().get_last_best_location():
#         best_location = read_packet.get_sensors().get_location().get_last_best_location()
#         location = StationLocation(best_location.get_latitude_longitude_timestamp().get_mach(),
#                                    best_location.get_altitude_timestamp().get_mach(),
#                                    best_location.get_speed_timestamp().get_mach(),
#                                    best_location.get_bearing_timestamp().get_mach(),
#                                    best_location.get_location_provider().name, best_location.get_score(),
#                                    best_location.get_latitude(), best_location.get_longitude(),
#                                    best_location.get_altitude(), best_location.get_speed(), best_location.get_bearing(),
#                                    best_location.get_horizontal_accuracy(), best_location.get_vertical_accuracy(),
#                                    best_location.get_speed_accuracy(), best_location.get_bearing_accuracy())
#     else:
#         location = None
#     metadata = StationMetadata(read_packet.get_station_information().get_id(),
#                                read_packet.get_station_information().get_make(),
#                                read_packet.get_station_information().get_model(), False,
#                                read_packet.get_station_information().get_os().name,
#                                read_packet.get_station_information().get_os_version(), "Redvox",
#                                read_packet.get_station_information().get_app_version(),
#                                read_packet.get_station_information().get_app_settings().get_scramble_audio_data(),
#                                timing, station_uuid=read_packet.get_station_information().get_uuid(),
#                                location_data=LocationData(location, [location]))
#     # add data from packets
#     data_dict = load_apim_wrapped_packet(read_packet)
#     timesync_data = TimeSyncData(np.array(read_packet.get_timing_information().get_synch_exchange_array()),
#                                  np.nan if read_packet.get_timing_information().get_best_latency() is None else
#                                  read_packet.get_timing_information().get_best_latency(),
#                                  0.0 if read_packet.get_timing_information().get_best_offset() is None else
#                                  read_packet.get_timing_information().get_best_offset())
#     packet_data = DataPacket(read_packet.get_timing_information().get_server_acquisition_arrival_timestamp(),
#                              read_packet.get_timing_information().get_app_start_mach_timestamp(),
#                              read_packet.get_timing_information().get_packet_start_mach_timestamp(),
#                              read_packet.get_timing_information().get_packet_end_mach_timestamp(),
#                              data_dict[SensorType.AUDIO].data_duration_s(),
#                              data_dict[SensorType.AUDIO].num_samples(), timesync_data)
#     packet_list: List[DataPacket] = [packet_data]
#     # get the best timing values for the station
#     if timing.station_best_latency is None or np.isnan(timing.station_best_latency):
#         ts_analysis = ts.TimeSyncData(packet_data, metadata)
#         metadata.timing_data.station_best_latency = ts_analysis.best_latency
#         metadata.timing_data.station_best_offset = ts_analysis.best_offset
#         metadata.timing_data.station_mean_offset = ts_analysis.mean_offset
#         metadata.timing_data.station_std_offset = ts_analysis.offset_std
#     return Station(metadata, data_dict, packet_list)
#
#
# def load_from_file_range_api_m(directory: str,
#                                start_timestamp_utc_s: Optional[int] = None,
#                                end_timestamp_utc_s: Optional[int] = None,
#                                redvox_ids: Optional[List[str]] = None,
#                                structured_layout: bool = False) -> ReadResult:
#     """
#     reads in api M data from a directory and returns a list of stations
#     :param directory: The root directory of the data. If structured_layout is False, then this directory will
#                       contain various unorganized .rdvxm files. If structured_layout is True, then this directory
#                       must be the root api1000 directory of the structured files.
#     :param start_timestamp_utc_s: The start timestamp in seconds since the epoch UTC.
#     :param end_timestamp_utc_s: The end timestamp in seconds since the epoch UTC.
#     :param redvox_ids: An optional list of redvox_ids to filter against, default empty list
#     :param structured_layout: An optional value to define if this is loading structured data, default False.
#     :return: a list of Station objects that contain the data
#     """
#     all_stations: ReadResult = ReadResult({})
#     all_data = apim_io.read_structured(directory, start_timestamp_utc_s, end_timestamp_utc_s, redvox_ids,
#                                        structured_layout)
#     for read_packets in all_data.all_wrapped_packets:
#         # set station metadata and timing based on first packet
#         if read_packets.wrapped_packets[0]:
#             if read_packets.wrapped_packets[0].get_sensors().get_audio():
#                 first_pack = read_packets.wrapped_packets[0]
#                 timing = StationTiming(read_packets.start_mach_timestamp, read_packets.audio_sample_rate,
#                                        first_pack.get_sensors().get_audio().get_first_sample_timestamp(),
#                                        start_timestamp_utc_s, end_timestamp_utc_s,
#                                        np.nan if first_pack.get_timing_information().get_best_latency() is None else
#                                        first_pack.get_timing_information().get_best_latency(),
#                                        0.0 if first_pack.get_timing_information().get_best_offset() is None else
#                                        first_pack.get_timing_information().get_best_offset())
#                 station_info = first_pack.get_station_information()
#                 metadata = StationMetadata(read_packets.redvox_id, station_info.get_make(), station_info.get_model(),
#                                            False, station_info.get_os().name, station_info.get_os_version(),
#                                            "Redvox", station_info.get_app_version(),
#                                            station_info.get_app_settings().get_scramble_audio_data(), timing,
#                                            station_uuid=read_packets.uuid)
#             else:
#                 raise ValueError("Error reading data window: Packet is missing Audio sensor!")
#         else:
#             raise ValueError("Error reading data window: First packet of data is missing!")
#         new_station = Station(metadata)
#         # add data from packets
#         packet_list: List[DataPacket] = []
#         for packet in read_packets.wrapped_packets:
#             data_dict = load_apim_wrapped_packet(packet)
#             new_station.append_station_data(data_dict)
#             best_latency = packet.get_timing_information().get_best_latency()
#             best_offset = packet.get_timing_information().get_best_offset()
#             timesync_data = TimeSyncData(np.array(packet.get_timing_information().get_synch_exchange_array()),
#                                          np.nan if best_latency is None else best_latency,
#                                          0.0 if best_offset is None else best_offset)
#             if packet.get_sensors().validate_location() and \
#                     packet.get_sensors().get_location().get_last_best_location():
#                 best_location = packet.get_sensors().get_location().get_last_best_location()
#                 location = StationLocation(best_location.get_latitude_longitude_timestamp().get_mach(),
#                                            best_location.get_altitude_timestamp().get_mach(),
#                                            best_location.get_speed_timestamp().get_mach(),
#                                            best_location.get_bearing_timestamp().get_mach(),
#                                            best_location.get_location_provider().name, best_location.get_score(),
#                                            best_location.get_latitude(), best_location.get_longitude(),
#                                            best_location.get_altitude(), best_location.get_speed(),
#                                            best_location.get_bearing(), best_location.get_horizontal_accuracy(),
#                                            best_location.get_vertical_accuracy(), best_location.get_speed_accuracy(),
#                                            best_location.get_bearing_accuracy())
#             else:
#                 location = None
#             packet_data = DataPacket(packet.get_timing_information().get_server_acquisition_arrival_timestamp(),
#                                      packet.get_timing_information().get_app_start_mach_timestamp(),
#                                      packet.get_timing_information().get_packet_start_mach_timestamp(),
#                                      packet.get_timing_information().get_packet_end_mach_timestamp(),
#                                      data_dict[SensorType.AUDIO].data_duration_s(),
#                                      data_dict[SensorType.AUDIO].num_samples(),
#                                      timesync_data, best_location=location)
#             packet_list.append(packet_data)
#         new_station.packet_data = packet_list
#
#         if timing.station_best_latency is None or np.isnan(timing.station_best_latency):
#             # get the best timing values for the station
#             ts_analysis = ts.TimeSyncAnalysis(new_station)
#             new_station.station_metadata.timing_data.station_best_latency = ts_analysis.get_best_latency()
#             new_station.station_metadata.timing_data.station_best_offset = ts_analysis.get_best_offset()
#             new_station.station_metadata.timing_data.station_mean_offset = ts_analysis.get_mean_offset()
#             new_station.station_metadata.timing_data.station_std_offset = ts_analysis.get_offset_stdev()
#
#         # create the Station data object
#         all_stations.append_station(f"{read_packets.redvox_id}:{read_packets.uuid}", new_station)
#     return all_stations
#
#
# # todo: drop support or refactor into api_m
# def load_from_mseed(file_path: str, station_ids: Optional[List[str]] = None) -> ReadResult:
#     """
#     load station data from a miniseed file
#     :param file_path: the location of the miniseed file
#     :param station_ids: the station ids to search for, default None; if None, get all stations
#     :return: a list of Station objects that contain the data
#     """
#     stations: ReadResult = ReadResult({})
#     strm = read(file_path)
#     for data_stream in strm:
#         record_info = data_stream.meta
#         start_time = int(dtu.seconds_to_microseconds(data_stream.meta["starttime"].timestamp))
#         end_time = int(dtu.seconds_to_microseconds(data_stream.meta["endtime"].timestamp))
#         station_timing = StationTiming(np.nan, record_info["sampling_rate"], start_time, start_time, end_time)
#         station_id = f'{record_info["network"]}{record_info["station"]}_{record_info["location"]}'
#         metadata = StationMetadata(station_id, "mb3_make", "mb3_model", False, "mb3_os", "mb3_os_vers",
#                                    "mb3_recorder", "mb3_recorder_version", False, station_timing,
#                                    record_info["calib"], record_info["network"], record_info["station"],
#                                    record_info["location"], record_info["channel"], record_info["mseed"]["encoding"])
#         sample_rate_hz = record_info["sampling_rate"]
#         timestamps = calc_evenly_sampled_timestamps(start_time, int(record_info["npts"]), sample_rate_hz)
#         data_for_df = np.transpose([timestamps, data_stream.data])
#         sensor_data = SensorData(record_info["channel"], pd.DataFrame(data_for_df, columns=["timestamps", "BDF"]),
#                                  record_info["sampling_rate"], 1 / record_info["sampling_rate"], 0.0, True)
#         data_packet = DataPacket(np.nan, start_time, start_time, end_time)
#         if station_ids is None or len(station_ids) == 0 or station_id in station_ids:
#             stations.append_station(f"{station_id}:{station_id}", Station(metadata, {SensorType.AUDIO: sensor_data},
#                                                                           [data_packet]))
#     return stations
#
#
# def read_all_in_dir(directory: str,
#                     start_timestamp_utc_s: Optional[int] = None,
#                     end_timestamp_utc_s: Optional[int] = None,
#                     station_ids: Optional[List[str]] = None,
#                     structured_layout: bool = False) -> ReadResult:
#     """
#     load all data files in the directory
#     :param directory: string, location of all the files;
#                         if structured_layout is True, the directory contains a root api1000, api900, or mseed directory,
#                         if structured_layout is False, the directory contains unsorted files
#     :param start_timestamp_utc_s: optional int, The start timestamp as seconds since the epoch UTC.
#     :param end_timestamp_utc_s: optional int, The end timestamp as seconds since the epoch UTC.
#     :param station_ids: optional list of string station ids to filter against, default empty list
#     :param structured_layout: optional bool to define if this is loading structured data, default False.
#     :return: a ReadResult object containing the data requested
#     """
#     # create the object to store the data
#     stations: ReadResult = ReadResult({})
#     # if structured_layout, there should be a specifically named folder in directory
#     if structured_layout:
#         if "api900" not in directory:
#             api900_dir = os.path.join(directory, "api900")
#         else:
#             api900_dir = directory
#         if "api1000" not in directory:
#             apim_dir = os.path.join(directory, "api1000")
#         else:
#             apim_dir = directory
#         if "mseed" not in directory:
#             mseed_dir = os.path.join(directory, "mseed")
#         else:
#             mseed_dir = directory
#         # check if none of the paths exists
#         if not (os.path.exists(api900_dir) or os.path.exists(apim_dir) or os.path.exists(mseed_dir)):
#             # no specially named directory found; raise error
#             raise ValueError(f"{directory} does not contain api900, api1000 or mseed directory.")
#     else:
#         # load files from unstructured layout; everything is sitting in the main directory
#         api900_dir = directory
#         apim_dir = directory
#         mseed_dir = directory
#
#     # get api900 data
#     stations.append(load_file_range_from_api900(api900_dir, start_timestamp_utc_s, end_timestamp_utc_s,
#                                                 station_ids, structured_layout, False))
#     # get api1000 data
#     stations.append(load_from_file_range_api_m(apim_dir, start_timestamp_utc_s, end_timestamp_utc_s,
#                                                station_ids, structured_layout))
#     # get mseed data
#     all_paths = glob.glob(os.path.join(mseed_dir, "*.mseed"))
#     for path in all_paths:
#         stations.append(load_from_mseed(path, station_ids))
#     return stations
